"""The implementation of the resampling methods for use with the Calendar."""
import typing
from typing import Callable
from typing import Literal
from typing import Union
from typing import overload
import numpy as np
import pandas as pd
import xarray as xr
from lilio.calendar import Calendar
from . import utils


# List of numpy statistical methods, with a single input argument and a single output.
ResamplingMethod = Literal[
    "mean",
    "min",
    "max",
    "median",
    "std",
    "var",
    "ptp",
    "nanmean",
    "nanmedian",
    "nanstd",
    "nanvar",
    "sum",
    "nansum",
    "size",
    "count_nonzero",
]
VALID_METHODS = typing.get_args(ResamplingMethod)


def _check_valid_resampling_methods(method: ResamplingMethod):
    """Check if a method is valid as a (numpy) resampling method."""
    if method not in VALID_METHODS:
        raise ValueError(
            f"{method} is not a valid resampling method. Choose from: {VALID_METHODS}"
        )


def _mark_target_period(
    input_data: Union[pd.DataFrame, xr.Dataset]
) -> Union[pd.DataFrame, xr.Dataset]:
    """Mark interval periods that fall within the given number of target periods.

    Pass a pandas Series/DataFrame with an 'i_interval' column, or an xarray
    DataArray/Dataset with an 'i_interval' coordinate axis. It will return an
    object with an added column in the Series/DataFrame or an
    added coordinate axis in the DataSet called 'target'. This is a boolean
    indicating whether the index time interval is a target period or not. This is
    determined by the instance variable 'n_targets'.

    Args:
        input_data: Input data for resampling. For a Pandas object, one of its
        columns must be called 'i_interval'. An xarray object requires a coordinate
        axis named 'i_interval' containing an interval counter for every period.

    Returns:
        Input data with boolean marked target periods, similar data format as
            given inputs.
    """
    if isinstance(input_data, (pd.Series, pd.DataFrame)):
        input_data["is_target"] = np.ones(input_data.index.size, dtype=bool)
        input_data["is_target"] = input_data["is_target"].where(
            input_data["i_interval"] > 0, other=False
        )

    else:
        # input data is xr.Dataset
        target = input_data["i_interval"] > 0
        input_data = input_data.assign_coords(coords={"is_target": target})

    return input_data


def _resample_bins_constructor(
    intervals: Union[pd.Series, pd.DataFrame]
) -> pd.DataFrame:
    """Restructures the interval object into a tidy DataFrame.

    Args:
        intervals: the output interval `pd.Series` or `pd.DataFrame` from the
            `map_to_data` function.

    Returns:
        Pandas DataFrame with 'anchor_year', 'i_interval', and 'interval' as
            columns.
    """
    # Make a tidy dataframe where the intervals are linked to the anchor year
    if isinstance(intervals, pd.DataFrame):
        bins = intervals.copy()
        bins.index.rename("anchor_year", inplace=True)
        bins = bins.melt(
            var_name="i_interval", value_name="interval", ignore_index=False
        )
        bins = bins.sort_values(by=["anchor_year", "i_interval"])
    else:
        # Massage the dataframe into the same tidy format for a single year
        bins = pd.DataFrame(intervals)
        bins = bins.melt(
            var_name="anchor_year", value_name="interval", ignore_index=False
        )
        bins.index.rename("i_interval", inplace=True)
    bins = bins.reset_index()

    return bins


def _contains(interval_index: pd.IntervalIndex, timestamps) -> np.ndarray:
    """Check elementwise if the intervals contain the timestamps.

    Will return a boolean array of the shape (n_timestamps, n_intervals).

    Args:
        interval_index: An IntervalIndex containing all intervals that should be checked.
        timestamps: A 1-D array containing

    Returns:
        np.ndarray: 2-D mask array
    """
    if interval_index.closed_left:
        a = np.greater_equal(timestamps, interval_index.left.values[np.newaxis].T)
    else:
        a = np.greater(timestamps, interval_index.left.values[np.newaxis].T)
    if interval_index.closed_right:
        b = np.less_equal(timestamps, interval_index.right.values[np.newaxis].T)
    else:
        b = np.less(timestamps, interval_index.right.values[np.newaxis].T)
    return a & b


def _resample_pandas(
    calendar: Calendar,
    input_data: Union[pd.Series, pd.DataFrame],
    how: Union[ResamplingMethod, Callable[[np.ndarray], np.ndarray]],
) -> pd.DataFrame:
    """Resample Pandas data.

    Args:
        calendar: Mapped Lilio Calendar object.
        input_data: Data provided by the user to the `resample` function
        how: Which resampling method should be used. Can also be a function that takes a
            single input argument and has a single output argument.

    Returns:
        pd.DataFrame: DataFrame containing the intervals and data resampled to
            these intervals.
    """
    resampling_method = getattr(np, how) if isinstance(how, str) else how

    if isinstance(input_data, pd.Series):
        name = "data" if input_data.name is None else input_data.name
        input_data = pd.DataFrame(input_data.rename(name))

    data = _resample_bins_constructor(calendar.get_intervals())
    contains_matrix = _contains(
        pd.IntervalIndex(data.interval.values), input_data.index.values
    )

    for colname in input_data.columns:
        resampled_data = np.zeros(contains_matrix.shape[0])
        for i, row in enumerate(contains_matrix):
            resampled_data[i] = resampling_method(input_data[colname].values[row])
        data[colname] = resampled_data

    return data


# pylint: disable=too-many-locals
def _resample_dataset(
    calendar: Calendar,
    input_data: xr.Dataset,
    how: Union[ResamplingMethod, Callable[[np.ndarray], np.ndarray]],
) -> xr.Dataset:
    """Resample xarray data.

    Args:
        calendar: A mapped Lilio Calendar object
        input_data (xr.DataArray or xr.Dataset): Data provided by the user to the
            `resample` function
        how: Which resampling method should be used. Can also be a function that takes a
            single input argument and has a single output argument.

    Returns:
        xr.Dataset: Dataset containing the intervals and data resampled to
            these intervals.
    """
    resampling_method = getattr(np, how) if isinstance(how, str) else how
    data = calendar.flat.to_xarray().rename("interval")
    data = data.to_dataset()
    data = data.stack(anch_int=("anchor_year", "i_interval"))

    # Separate data with time dims (should be resampled), from data without time dims
    #   (which does not need resampling). Otherwise stacking ALL dims together will
    #   cause the stacked dimension become needlessly large, making resampling slow.
    input_data_time = input_data[
        [var for var in input_data.data_vars if "time" in input_data[var].dims]
    ]
    input_data_nontime = input_data[
        [var for var in input_data.data_vars if "time" not in input_data[var].dims]
    ]

    stacking_dims = list(input_data_time.dims.keys())
    stacking_dims.remove("time")
    if stacking_dims:  # There might not be extra dims to stack!
        input_data_time = input_data_time.stack(allstack=stacking_dims)

    da_coords = {"anch_int": data["anch_int"]}
    if stacking_dims:
        da_coords["allstack"] = input_data_time["allstack"]

    contains_matrix = _contains(
        pd.IntervalIndex(data["interval"].values), input_data_time["time"].values
    )

    resampled_vars = [xr.DataArray] * len(input_data_time.data_vars)
    for i, var in enumerate(input_data_time.data_vars):
        size_input = input_data_time[var].shape[1] if stacking_dims else 1
        resampled_data = np.zeros((contains_matrix.shape[0], size_input))

        # Note: while looping, contains_matrix will have a size of
        # (n_intervals * n_anchor_years), making it the most reliably small dim.
        _data = input_data_time[var].values
        for j, row in enumerate(contains_matrix):
            resampled_data[j, :] = resampling_method(_data[row], axis=0)

        resampled_data = np.squeeze(resampled_data)  # in case of (1, n) resampled data

        resampled_vars[i] = xr.DataArray(  # type: ignore
            data=resampled_data, coords=da_coords
        ).rename(var)

    if input_data_nontime.data_vars:
        data = xr.merge([data, input_data_nontime] + resampled_vars)
    else:
        data = xr.merge([data] + resampled_vars)

    data = data.unstack()
    data = utils.convert_interval_to_bounds(data)
    data = data.transpose("anchor_year", "i_interval", ...)
    return data.sortby("anchor_year", "i_interval")


@overload
def resample(calendar: Calendar, input_data: xr.Dataset) -> xr.Dataset:
    ...


@overload
def resample(calendar: Calendar, input_data: xr.DataArray) -> xr.DataArray:
    ...


@overload
def resample(
    calendar: Calendar, input_data: Union[pd.Series, pd.DataFrame]
) -> pd.DataFrame:
    ...


def resample(
    calendar: Calendar,
    input_data: Union[pd.Series, pd.DataFrame, xr.DataArray, xr.Dataset],
    how: Union[ResamplingMethod, Callable[[np.ndarray], np.ndarray]] = "mean",
) -> Union[pd.DataFrame, xr.DataArray, xr.Dataset]:
    """Resample input data to the Calendar's intervals.

    Pass a pandas Series/DataFrame with a datetime axis, or an
    xarray DataArray/Dataset with a datetime coordinate called 'time'.
    It will return the same object with the datetimes resampled onto
    the Calendar's Index by binning the data into the Calendar's intervals
    and calculating the mean of each bin.

    The default behavior is to calculate the mean for each interval. However, many other
    statistics can be calculated, namely:
        - mean, min, max, median
        - std, var, ptp (peak-to-peak)
        - nanmean, nanmedian, nanstd, nanvar
        - sum, nansum, size, count_nonzero

    "size" will compute the number of datapoints that are in each interval, and can be
    used to check if the input data is of a sufficiently high resolution to be resampled
    to the calendar.

    Note: this function is intended for upscaling operations, which means
    the calendar frequency is larger than the original frequency of input data (e.g.
    `freq` is "7days" and the input is daily data). It supports downscaling
    operations but the user need to be careful since the returned values may contain
    "NaN".

    Args:
        calendar: Calendar object with either a map_year or map_to_data mapping.
        input_data: Input data for resampling. For a Pandas object its index must be
            either a pandas.DatetimeIndex. An xarray object requires a dimension
            named 'time' containing datetime values.
        how: Which method for resampling should be used. Either a string or function.
            The following methods are supported as a string input:
                mean, min, max, median
                std, var, ptp (peak-to-peak)
                nanmean, nanmedian, nanstd, nanvar
                sum, nansum, size, count_nonzero
            Alternatively, a function can be passed. For example
            `resample(how=np.mean)`.

    Raises:
        UserWarning: If the calendar frequency is smaller than the frequency of
            input data

    Returns:
        Input data resampled based on the calendar frequency, similar data format as
            given inputs.

    Example:
        Assuming the input data is pd.DataFrame containing random values with index
        from 2021-11-11 to 2021-11-01 at daily frequency.

        >>> import lilio
        >>> import pandas as pd
        >>> import numpy as np
        >>> cal = lilio.daily_calendar(anchor="12-31", length="180d")
        >>> time_index = pd.date_range("20191201", "20211231", freq="1d")
        >>> var = np.arange(len(time_index))
        >>> input_data = pd.Series(var, index=time_index)
        >>> cal = cal.map_to_data(input_data)
        >>> bins = lilio.resample(cal, input_data)
        >>> bins # doctest: +NORMALIZE_WHITESPACE
            anchor_year  i_interval                  interval   data  is_target
        0          2019          -1  [2019-07-04, 2019-12-31)   14.5      False
        1          2019           1  [2019-12-31, 2020-06-28)  119.5       True
        2          2020          -1  [2020-07-04, 2020-12-31)  305.5      False
        3          2020           1  [2020-12-31, 2021-06-29)  485.5       True
    """
    if calendar.mapping is None:
        raise ValueError("Generate a calendar map before calling resample")

    if isinstance(how, str):
        _check_valid_resampling_methods(how)
    utils.check_timeseries(input_data)
    utils.check_input_frequency(calendar, input_data)
    utils.check_reserved_names(input_data)

    if isinstance(input_data, (pd.Series, pd.DataFrame)):
        resampled_data = _resample_pandas(calendar, input_data, how)
    elif isinstance(input_data, xr.DataArray):
        da_name = "data" if input_data.name is None else input_data.name
        input_data.name = da_name
        resampled_data = _resample_dataset(calendar, input_data.to_dataset(), how)
    else:
        resampled_data = _resample_dataset(calendar, input_data, how)

    utils.check_empty_intervals(resampled_data)

    resampled_data = _mark_target_period(resampled_data)

    if isinstance(input_data, xr.DataArray):
        return resampled_data[da_name]
    return resampled_data



"""Calendar shifter to create staggered calendars."""
import copy
from typing import Dict
from typing import List
from typing import Union
import xarray as xr
from . import calendar
from . import utils
from .resampling import resample


def _gap_shift(
    interval: calendar.Interval, shift: Union[str, Dict[str, int]]
) -> Dict[str, int]:
    """Shift a calendar interval's gap property by the given amount.

    Args:
        interval: The interval that will be shifted
        gap: the pandas DateOffset from a calendar interval's `gap` property.
        shift: the shift for the gap, in the form of a pandas-like frequency
            string (e.g. "10d", "2W", or "3M"), or a pandas.DateOffset compatible
            dictionary such as {days=10}, {weeks=2}, or {months=1, weeks=2}.

    Returns:
        A Pandas DateOffset compatible dictionary, with the gap offset by shift
    """
    if isinstance(interval.gap, str):
        gap_time_dict = utils.parse_freqstr_to_dateoffset(interval.gap)
    else:
        gap_time_dict = interval.gap

    if isinstance(shift, str):
        shift_time_dict = utils.parse_freqstr_to_dateoffset(shift)
    else:
        shift_time_dict = shift
    # make the shift negative for the precursor to shift forward in time
    if interval.role == "precursor":
        shift_time_dict.update(
            (key, value * -1) for key, value in shift_time_dict.items()
        )
    return {
        k: gap_time_dict.get(k, 0) + shift_time_dict.get(k, 0)
        for k in set(gap_time_dict) | set(shift_time_dict)
    }


def calendar_shifter(
    calendar: calendar.Calendar, shift: Union[str, dict]
) -> calendar.Calendar:
    """Shift a Calendar instance by a given time offset.

    Instead of shifting the anchor date, this function shifts two things in reference
    to the anchor date:
        target period(s): as a gap between the anchor date and the start of the first
            target period
        precursor period(s): as a gap between the anchor date and the start of the
            first precursor period
    This way, the anchor year from the input calendar is maintained on the returned
    calendar. This is important for train-test splitting at later stages.

    Args:
        calendar: a lilio.Calendar instance
        shift: a pandas-like
            frequency string (e.g. "10d", "2W", or "3M"), or a pandas.DateOffset
            compatible dictionary such as {days=10}, {weeks=2}, or {months=1, weeks=2}

    Example:
        Shift a calendar by a given dateoffset.

        >>> import lilio
        >>> cal = lilio.Calendar(anchor="07-01")
        >>> cal.add_intervals("target", "7d")
        >>> cal.add_intervals("precursor", "7d", gap="14d")
        >>> cal.add_intervals("precursor", "7d", n=3)
        >>> cal_shifted = lilio.calendar_shifter.calendar_shifter(cal, "7d")
        >>> cal_shifted  # doctest: +NORMALIZE_WHITESPACE
        Calendar(
            anchor='07-01',
            allow_overlap=False,
            mapping=None,
            intervals=[
                Interval(role='target', length='7d', gap={'days': 7}),
                Interval(role='precursor', length='7d', gap={'days': 7}),
                Interval(role='precursor', length='7d', gap='0d'),
                Interval(role='precursor', length='7d', gap='0d'),
                Interval(role='precursor', length='7d', gap='0d')
            ]
        )
    """
    calendar_shifted = copy.deepcopy(calendar)
    calendar_shifted.targets[0].gap = _gap_shift(calendar.targets[0], shift)
    calendar_shifted.precursors[0].gap = _gap_shift(calendar.precursors[0], shift)

    return calendar_shifted


def staggered_calendar(
    calendar: calendar.Calendar, shift: Union[str, dict], n_shifts: int
) -> List[calendar.Calendar]:
    """Create a staggered calendar list by shifting a calendar by an offset n-times.

    Args:
        calendar: an lilio.Calendar instance
        shift: a pandas-like
            frequency string (e.g. "10d", "2W", or "3M"), or a pandas.DateOffset
            compatible dictionary such as {days=10}, {weeks=2}, or {months=1, weeks=2}
        n_shifts: strictly positive integer for the number of shifts

    Example:
        Shift an input calendar n times by a given dateoffset and return a list of these
        shifted calendars.

        >>> import lilio
        >>> cal = lilio.Calendar(anchor="07-01")
        >>> cal.add_intervals("target", "7d")
        >>> cal.add_intervals("precursor", "7d", gap="14d")
        >>> cal.add_intervals("precursor", "7d", n=3)
        >>> cal_shifted = lilio.calendar_shifter.staggered_calendar(cal, "7d", 1)
        >>> cal_shifted  # doctest: +NORMALIZE_WHITESPACE
        [Calendar(
            anchor='07-01',
            allow_overlap=False,
            mapping=None,
            intervals=[
                Interval(role='target', length='7d', gap='0d'),
                Interval(role='precursor', length='7d', gap='14d'),
                Interval(role='precursor', length='7d', gap='0d'),
                Interval(role='precursor', length='7d', gap='0d'),
                Interval(role='precursor', length='7d', gap='0d')
            ]
        ),
        Calendar(
            anchor='07-01',
            allow_overlap=False,
            mapping=None,
            intervals=[
                Interval(role='target', length='7d', gap={'days': 7}),
                Interval(role='precursor', length='7d', gap={'days': 7}),
                Interval(role='precursor', length='7d', gap='0d'),
                Interval(role='precursor', length='7d', gap='0d'),
                Interval(role='precursor', length='7d', gap='0d')
            ]
        )]
    """
    if not isinstance(n_shifts, int):
        raise ValueError(
            "Please input an 'int' type for the 'n' argument."
            f" Not a {type(n_shifts)}."
        )
    if n_shifts <= 0:
        raise ValueError(
            "The number of shifts 'n' has to be 1 or greater, " f"not '{n_shifts}'."
        )
    cal_staggered = [calendar]
    for _ in range(n_shifts):
        cal_shifted = calendar_shifter(cal_staggered[-1], shift)
        cal_staggered.append(cal_shifted)

    return cal_staggered


def calendar_list_resampler(
    cal_list: list, ds: xr.Dataset, dim_name: str = "step"
) -> xr.Dataset:
    """Return a dataset, resampled to every calendar in a list of calendars.

    The resampled calendars will be concatenated along a new dimension (with the default
    name 'step') into a single xarray Dataset.

    Args:
        cal_list: List of calendars.
        ds: Dataset to resample.
        dim_name: The name of the new dimension that will be added to the output
            dataset.

    Returns:
        Resampled xr.Dataset
    """
    ds_r = xr.concat([resample(cal, ds) for cal in cal_list], dim=dim_name)
    ds_r = ds_r.assign_coords({dim_name: ds_r[dim_name].values})

    return ds_r



"""Tests for the lilio._bokeh_plots module."""
import matplotlib as mpl
import matplotlib.pyplot as plt
import pytest
from bokeh import io as bokeh_io
import lilio


mpl.use("Agg")  # required for windows


class TestPlots:
    """Test the visualizations (rough check for errors only)."""

    @pytest.fixture
    def dummy_bokeh_file(self, tmp_path):
        bokeh_io.output_file(tmp_path / "test.html")

    custom_cal_pre = lilio.Calendar(anchor="12-31")
    custom_cal_pre.add_intervals("precursor", "10d")
    custom_cal_tar = lilio.Calendar(anchor="12-31")
    custom_cal_tar.add_intervals("target", "10d")

    calendars = [
        lilio.daily_calendar(anchor="12-31", length="60d"),
        lilio.monthly_calendar(anchor="December", length="1M"),
        lilio.weekly_calendar(anchor="W40", length="2W"),
        custom_cal_pre,
        custom_cal_tar,
    ]

    @pytest.fixture(params=calendars, autouse=True)
    def dummy_calendars(self, request):
        "Dummy that tests all available calendars."
        cal = request.param
        return cal.map_years(2018, 2021)

    @pytest.fixture(params=[True, False], autouse=True)
    def isinteractive(self, request):
        return request.param

    def test_visualize_relative(self, dummy_calendars, isinteractive):
        dummy_calendars.visualize(interactive=isinteractive, relative_dates=True)
        plt.close("all")

    def test_visualize_absolute(self, dummy_calendars, isinteractive):
        dummy_calendars.visualize(interactive=isinteractive, relative_dates=False)
        plt.close("all")

    def test_visualize_without_text(self, dummy_calendars, isinteractive):
        dummy_calendars.visualize(interactive=isinteractive, show_length=False)
        plt.close("all")

    def test_visualize_with_text(self, dummy_calendars, isinteractive):
        dummy_calendars.visualize(interactive=isinteractive, show_length=True)
        plt.close("all")

    def test_visualize_unmapped(self, isinteractive):
        lilio.daily_calendar(anchor="12-31").visualize(interactive=isinteractive)
        plt.close("all")


class TestPlotsSingle:
    """Test the visualization routines, where a single calendar suffices"""

    @pytest.fixture
    def dummy_bokeh_file(self, tmp_path):
        bokeh_io.output_file(tmp_path / "test.html")

    @pytest.fixture
    def dummy_calendar(self):
        "Dummy that will only test for daily_calendar (to avoid excess testing)"
        cal = lilio.daily_calendar(anchor="12-31", length="60d")
        return cal.map_years(2018, 2021)

    def test_bokeh_kwargs(self, dummy_calendar):
        """Testing kwargs that overwrite default kwargs."""
        dummy_calendar.visualize(
            interactive=True,
            width=800,
            height=200,
            tooltips=[("Interval", "@desc")],
        )

    def test_bokeh_kwarg_mpl(self, dummy_calendar):
        with pytest.warns():
            dummy_calendar.visualize(interactive=False, width=800)
            plt.close("all")

    def test_mpl_no_legend(self, dummy_calendar):
        dummy_calendar.visualize(interactive=False, add_legend=False)
        plt.close("all")

    def test_mpl_ax_kwarg(self, dummy_calendar):
        _, ax = plt.subplots()
        dummy_calendar.visualize(interactive=False, ax=ax)
        plt.close("all")

    def test_bokeh_ax_warning(self, dummy_calendar):
        _, ax = plt.subplots()
        with pytest.warns():
            dummy_calendar.visualize(interactive=True, ax=ax)



"""Calendar plotting implementations (general and matplotlib)."""
import typing
from typing import Dict
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import dates as mdates
from matplotlib.patches import Patch
from matplotlib.patches import Rectangle


if typing.TYPE_CHECKING:
    from lilio.calendar import Calendar


def make_color_array(n_targets: int, n_intervals: int) -> np.ndarray:
    """Util that generates the colormap for the intervals.

    Args:
        n_targets: number of target intervals
        n_intervals: total number of intervals

    Returns:
        1-D array containing hex colors.
    """
    colors = np.array(["#ffffff"] * n_intervals)
    if n_targets > 0:
        colors[0:n_targets:2] = "#ff7700"  # Orange
        colors[1:n_targets:2] = "#ffa100"  # Orange (darker)
    if n_intervals - n_targets > 0:
        colors[n_targets::2] = "#1f9ce9"  # Blue
        colors[n_targets + 1 :: 2] = "#137fc1"  # Blue (darker)
    return colors[::-1]


def _get_xdata(
    relative_dates: bool,
    year_intervals: pd.Series,
    widths: np.ndarray,
    anchor_date: pd.Timestamp,
) -> np.ndarray:
    """Util that generates the x-coordinate of every interval rectangle for the plot.

    Args:
        relative_dates: If False, absolute dates will be used. If True, each anchor year
                        is aligned by the anchor date, so that all anchor years line up
                        vertically.
        year_intervals: The Series of intervals for a single anchor year.
        widths: Width of each interval, as generated by _get_widths.
        anchor_date: The anchor date.

    Returns:
        1-D array containing the x-coordinates for every interval rectangle.
    """
    if relative_dates:
        x_data = np.array([(i.left - anchor_date).days for i in year_intervals])
        return x_data + 0.5 * widths

    x_data = np.array([i.left for i in year_intervals])
    return x_data + 0.5 * widths


def _get_widths(
    relative_dates: bool,
    year_intervals: pd.Series,
) -> np.ndarray:
    """Util that generates the width of every interval rectangle for the plot.

    Args:
        relative_dates: If False, absolute dates will be used. If True, each anchor year
                        is aligned by the anchor date, so that all anchor years line up
                        vertically.
        year_intervals: The series of intervals for a single anchor year.

    Returns:
        1-D array containing the width of every interval rectangle.
    """
    if relative_dates:
        return np.array([(i.right - i.left).days for i in year_intervals])
    return np.array([(i.right - i.left) for i in year_intervals])


def generate_plot_data(
    calendar: "Calendar",
    relative_dates: bool,
    year: int,
    year_intervals: pd.Series,
) -> Dict:
    """Util to generate the plotting data, containing all variables to plot.

    Args:
        calendar: The calendar you want to plot.
        relative_dates: If False, absolute dates will be used. If True, each anchor year
                        is aligned by the anchor date, so that all anchor years line up
                        vertically.
        year: The anchor year for which plot data should be generated.
        year_intervals: The pandas series of intervals for a single anchor year.

    Returns:
        Dict: Dictionary containing all the data to generate the Matplotlib or Bokeh
            plots.
    """
    n_targets = calendar.n_targets

    anchor_date = calendar._get_anchor(year=year)  # pylint: disable=protected-access

    widths = _get_widths(relative_dates, year_intervals)
    interval_str = np.array(
        [f"{str(i.left)[:10]} -> {str(i.right)[:10]}" for i in year_intervals]
    )

    types = np.array(["Precursor"] * len(year_intervals))
    types[-n_targets:] = "Target"

    width_days = widths if relative_dates else np.array([x.days for x in widths])

    return {
        "x": _get_xdata(relative_dates, year_intervals, widths, anchor_date),
        "y": np.ones(len(year_intervals)) * anchor_date.year,
        "height": np.ones(len(year_intervals)) * 0.8,
        "width": widths,
        "width_days": width_days,
        "color": make_color_array(n_targets, len(year_intervals)),
        "desc": interval_str,
        "type": types,
    }


def plot_rectangles(ax: plt.Axes, data: Dict, show_length: bool):
    """Generate rectangles from the input data.

    Args:
        ax: Axis in which the rectangles should be shown.
        data: Data dictionary containing x, y, width, height, color and width_days.
        show_length: If the length of each interval should be displayed as text.
    """
    for _, row in pd.DataFrame(data).iterrows():  # type: ignore
        ax.add_patch(
            Rectangle(
                xy=(row["x"] - row["width"] / 2, row["y"] - row["height"] / 2),
                width=row["width"],
                height=row["height"],
                facecolor=row["color"],
                alpha=0.7,
                edgecolor="k",
                linewidth=1.5,
            )
        )

        if show_length:
            ax.text(
                x=row["x"],
                y=row["y"],
                s=f"{row['width_days']}",
                c="k",
                size=8,
                ha="center",
                va="center",
            )


def matplotlib_visualization(  # noqa: PLR0913 (too-many-arguments)
    calendar,
    n_years: int,
    relative_dates: bool,
    show_length: bool = False,
    add_legend: bool = True,
    add_yticklabels: bool = True,
    ax=None,
):
    """Visualization routine for generating a calendar visualization with Matplotlib.

    Args:
        calendar: Mapped calendar which should be visualized.
        n_years: Number of years which should be displayed (most recent years only).
        relative_dates: If False, absolute dates will be used. If True, each anchor year
                        is aligned by the anchor date, so that all anchor years line up
                        vertically.
        show_length: If the length of every periods should be displayed. Defaults False.
        add_legend: Add a legend to the plot. Defaults to True.
        add_yticklabels: If the years should be displayed on the y-axis ticks.
        ax: If a matplotlib axis object is passed, the plot will be generated in that
            axis. Otherwise a new figure is generated.
    """
    if ax is None:
        fig, ax = plt.subplots(figsize=(7, 4))
    else:
        fig = None

    intervals = calendar.get_intervals()[:n_years]

    for year, year_intervals in intervals.iterrows():
        data = generate_plot_data(
            calendar=calendar,
            relative_dates=relative_dates,
            year=year,
            year_intervals=year_intervals,
        )
        plot_rectangles(ax, data, show_length)

    ax.set_xlabel("Days before anchor date" if relative_dates else "Date")
    ax.set_ylabel("Anchor year")

    if relative_dates:
        ax.set_xlim(
            (
                np.min(data["x"]) - data["width"][np.argmin(data["x"])] / 2 - 5,  # type: ignore
                np.max(data["x"]) + data["width"][np.argmax(data["x"])] / 2 + 5,  # type: ignore
            )
        )
    else:
        formatter = mdates.DateFormatter("%Y-%m-%d")
        ax.xaxis.set_major_formatter(formatter)
        ax.set_xlim(
            (
                intervals.stack().values.min().left - pd.Timedelta(days=2),
                intervals.stack().values.max().right + pd.Timedelta(days=2),
            )
        )

    ax.set_ylim([intervals.index.min() - 0.5, intervals.index.max() + 0.5])  # type: ignore
    ax.set_yticks([int(x) for x in intervals.index.to_list()])

    if not add_yticklabels:
        ax.set_yticklabels([])

    # Add a custom legend to explain to users what the colors mean
    if add_legend:
        legend_elements = [
            Patch(
                facecolor="#ff8c00",
                label="Target interval",
                linewidth=1.5,
            ),
            Patch(
                facecolor="#137fc1",
                label="Precursor interval",
                linewidth=1.5,
            ),
        ]
        ax.legend(handles=legend_elements, loc="center left", bbox_to_anchor=(1, 0.5))

    if not relative_dates and fig is not None:
        fig.autofmt_xdate()  # Formats the dates nicely so they do not overlap



"""Calendar plotting implementations (bokeh specific code)."""
import sys
import typing
import numpy as np
from bokeh import plotting
from ._plot import generate_plot_data


if typing.TYPE_CHECKING:
    from lilio.calendar import Calendar


def _generate_rectangle(figure: plotting.Figure, source: plotting.ColumnDataSource):
    """Add intervals to the figure as rectangles.

    Args:
        figure: Bokeh figure in which the rectangles should be added.
        source: Bokeh source data containing the required parameters.

    Returns:
        plotting.figure
    """
    return figure.rect(
        x="x",
        y="y",
        width="width",
        height="height",
        line_color="#000000",
        fill_color="color",
        fill_alpha=0.7,
        line_width=1.5,
        source=source,
    )


def _bokeh_visualization(
    calendar: "Calendar",
    n_years: int,
    relative_dates: bool,
    add_yticklabels: bool = True,
    **kwargs,
) -> plotting.Figure:
    """Visualization routine for generating a calendar visualization with Bokeh.

    Args:
        calendar: Mapped calendar which should be visualized.
        n_years: Number of years which should be displayed (most recent years only).
        relative_dates: If False, absolute dates will be used. If True, each anchor year
                        is aligned by the anchor date, so that all anchor years line up
                        vertically.
        add_yticklabels: If the years should be displayed on the y-axis ticks.
        **kwargs: Keyword arguments that should be passed to Bokeh's plotting.figure.

    Returns:
        plotting.Figure
    """
    if add_yticklabels:
        tooltips = [
            ("Interval", "@desc"),
            ("Size", "@width_days days"),
            ("Type", "@type"),
        ]
    else:
        # Do not show the actual intervals, as the calendar is not mapped.
        tooltips = [("Size", "@width_days days"), ("Type", "@type")]

    if "width" not in kwargs:
        kwargs["width"] = 500
    if "height" not in kwargs:
        kwargs["height"] = 300
    if "tooltips" not in kwargs:
        kwargs["tooltips"] = tooltips
    if "x_axis_type" not in kwargs:
        kwargs["x_axis_type"] = "linear" if relative_dates else "datetime"

    figure = plotting.figure(
        **kwargs,
    )

    intervals = calendar.get_intervals()[:n_years]

    for year, year_intervals in intervals.iterrows():
        data = generate_plot_data(
            calendar=calendar,
            relative_dates=relative_dates,
            year=year,
            year_intervals=year_intervals,
        )
        _generate_rectangle(figure, plotting.ColumnDataSource(data))

    figure.xaxis.axis_label = (
        "Days relative to anchor date" if relative_dates else "Date"
    )
    figure.yaxis.axis_label = "Anchor year"

    if relative_dates:
        figure.x_range.start = (
            np.min(data["x"]) - data["width"][np.argmin(data["x"])] / 2 - 14  # type: ignore
        )
        figure.x_range.end = (
            np.max(data["x"]) + data["width"][np.argmax(data["x"])] / 2 + 14  # type: ignore
        )

    if add_yticklabels:
        figure.yaxis.ticker = [int(x) for x in intervals.index.to_list()]
    else:
        figure.yaxis.ticker = []

    return figure


def bokeh_visualization(
    calendar,
    n_years: int,
    relative_dates: bool,
    add_yticklabels: bool = True,
    **kwargs,
) -> None:
    """Visualization routine for generating a calendar visualization with Bokeh.

    Args:
        calendar: Mapped calendar which should be visualized.
        n_years: Number of years which should be displayed (most recent years only).
        relative_dates: If False, absolute dates will be used. If True, each anchor year
                        is aligned by the anchor date, so that all anchor years line up
                        vertically.
        add_yticklabels: If the years should be displayed on the y-axis ticks.
        **kwargs: Any other kwargs that should be passed to bokeh.plotting.figure()
    """
    bokeh_fig = _bokeh_visualization(
        calendar, n_years, relative_dates, add_yticklabels, **kwargs
    )
    if "pytest" in sys.modules:  # Do not open browser if we are testing.
        plotting.save(bokeh_fig)
    else:
        plotting.show(bokeh_fig)



"""Tests for the s2s.nwp_calendar module."""
import copy
import numpy as np
import pandas as pd
import pytest
import xarray as xr
from lilio import Calendar
from lilio import calendar_shifter


class TestCalendarShifter:
    """Test the calendar shifter methods."""

    @pytest.fixture
    def dummy_calendar(self):
        cal = Calendar(anchor="12-31")
        cal.add_intervals("target", "7d")
        cal.add_intervals("precursor", "7d", "7d")
        cal.add_intervals("precursor", "7d")
        return cal

    @pytest.fixture
    def dummy_data(self):
        time_index = pd.date_range("20211001", "20220501", freq="1d")
        np.random.seed(0)
        temperature = 15 + 8 * np.random.randn(2, 2, len(time_index))
        lon = [[-99.83, -99.32], [-99.79, -99.23]]
        lat = [[42.25, 42.21], [42.63, 42.59]]
        return xr.Dataset(
            data_vars={
                "temperature": (["x", "y", "time"], temperature),
            },
            coords={
                "lon": (["x", "y"], lon),
                "lat": (["x", "y"], lat),
                "time": time_index,
            },
            attrs={"description": "Weather related data."},
        )

    def test_calendar_shifter(self, dummy_calendar):
        shift = {"days": 7}
        cal_shifted = calendar_shifter.calendar_shifter(dummy_calendar, shift)
        cal_expected = copy.deepcopy(dummy_calendar)
        cal_expected.targets[0].gap = {"days": 7}
        cal_expected.precursors[0].gap = {"days": 0}
        assert len(cal_shifted.targets + cal_shifted.precursors) == len(
            cal_expected.targets + cal_expected.precursors
        )
        assert str(cal_shifted.targets) == str(cal_expected.targets)
        assert str(cal_expected.precursors) == str(cal_shifted.precursors)

    def test_staggered_calendar(self, dummy_calendar):
        cal_list = calendar_shifter.staggered_calendar(dummy_calendar, "7d", n_shifts=3)
        assert len(cal_list) == 4

    def test_calendar_list_resampler(self, dummy_calendar, dummy_data):
        cal_list = calendar_shifter.staggered_calendar(
            dummy_calendar, "14d", n_shifts=5
        )
        for cal in cal_list:
            cal.map_to_data(dummy_data)
        ds_resampled = calendar_shifter.calendar_list_resampler(cal_list, dummy_data)
        expected_coords = np.array([0, 1, 2, 3, 4, 5])
        assert np.array_equal(ds_resampled.step.values, expected_coords)



"""lilio: Calendar generator for machine learning with timeseries data.

Time indices are anchored to the target period of interest. By keeping
observations from the same cycle (typically 1 year) together and paying close
attention to the treatment of adjacent cycles, we avoid information leakage
between train and test sets.

Example:
    Countdown the 4 weeks until New Year's Eve

    >>> import lilio
    >>> calendar = lilio.Calendar(anchor="12-31")
    >>> calendar.add_intervals("target", "1d")
    >>> calendar.add_intervals("precursor", "4W", n=4)
    >>> calendar # doctest: +NORMALIZE_WHITESPACE
    Calendar(
        anchor='12-31',
        allow_overlap=False,
        mapping=None,
        intervals=[
            Interval(role='target', length='1d', gap='0d'),
            Interval(role='precursor', length='4W', gap='0d'),
            Interval(role='precursor', length='4W', gap='0d'),
            Interval(role='precursor', length='4W', gap='0d'),
            Interval(role='precursor', length='4W', gap='0d')
        ]
    )

    Get the 180-day periods leading up to New Year's eve for the year 2020

    >>> calendar = lilio.daily_calendar(anchor="12-31", length="180d")
    >>> calendar = calendar.map_years(2020, 2020)
    >>> calendar.show() # doctest: +NORMALIZE_WHITESPACE
    i_interval                         -1                         1
    anchor_year
    2020         [2020-07-04, 2020-12-31)  [2020-12-31, 2021-06-29)

    Get the 180-day periods leading up to New Year's eve for 2020 - 2022 inclusive.

    >>> calendar = lilio.daily_calendar(anchor="12-31", length="180d")
    >>> calendar = calendar.map_years(2020, 2022)
    >>> # note the leap year:
    >>> calendar.show() # doctest: +NORMALIZE_WHITESPACE
    i_interval                         -1                         1
    anchor_year
    2022         [2022-07-04, 2022-12-31)  [2022-12-31, 2023-06-29)
    2021         [2021-07-04, 2021-12-31)  [2021-12-31, 2022-06-29)
    2020         [2020-07-04, 2020-12-31)  [2020-12-31, 2021-06-29)

    To get a stacked representation:

    >>> calendar.map_years(2020, 2022).flat
    anchor_year  i_interval
    2022         -1            [2022-07-04, 2022-12-31)
                  1            [2022-12-31, 2023-06-29)
    2021         -1            [2021-07-04, 2021-12-31)
                  1            [2021-12-31, 2022-06-29)
    2020         -1            [2020-07-04, 2020-12-31)
                  1            [2020-12-31, 2021-06-29)
    dtype: interval

"""

import logging
from . import calendar_shifter
from . import traintest
from .calendar import Calendar
from .calendar import Interval
from .calendar_shorthands import daily_calendar
from .calendar_shorthands import monthly_calendar
from .calendar_shorthands import weekly_calendar
from .resampling import resample


logging.getLogger(__name__).addHandler(logging.NullHandler())

__author__ = "Yang Liu"
__email__ = "y.liu@esciencecenter.nl"
__version__ = "0.3.1"

__all__ = [
    "Calendar",
    "Interval",
    "resample",
    "daily_calendar",
    "monthly_calendar",
    "weekly_calendar",
    "calendar_shifter",
    "traintest",
]



"""Tests for Lilio's traintest module."""
import numpy as np
import pandas as pd
import pytest
import xarray as xr
from sklearn.model_selection import KFold
import lilio
import lilio.traintest


@pytest.fixture
def dummy_data():
    # Generate random data
    n = 50
    time_index = pd.date_range("20151020", periods=n, freq="60d")
    time_coord = {"time": time_index}
    x1 = xr.DataArray(np.random.randn(n), coords=time_coord, name="x1")
    x2 = xr.DataArray(np.random.randn(n), coords=time_coord, name="x2")
    y = xr.DataArray(np.random.randn(n), coords=time_coord, name="y")

    # Map data to calendar and store for later reference
    calendar = lilio.daily_calendar(anchor="10-15", length="180d")
    calendar.map_to_data(x1)
    x1 = lilio.resample(calendar, x1)
    print(x1)
    x2 = lilio.resample(calendar, x2)
    y = lilio.resample(calendar, y)
    return x1, x2, y


def test_kfold_x(dummy_data):
    """Correctly split x."""
    x1, _, _ = dummy_data
    cv = lilio.traintest.TrainTestSplit(KFold(n_splits=3))
    x_train, x_test = next(cv.split(x1))
    expected_train = [2019, 2020, 2021, 2022]
    expected_test = [2016, 2017, 2018]
    assert np.array_equal(x_train.anchor_year, expected_train)
    xr.testing.assert_equal(x_test, x1.sel(anchor_year=expected_test))


def test_kfold_xy(dummy_data):
    """Correctly split x and y."""
    x1, _, y = dummy_data
    cv = lilio.traintest.TrainTestSplit(KFold(n_splits=3))
    x_train, x_test, y_train, y_test = next(cv.split(x1, y=y))
    expected_train = [2019, 2020, 2021, 2022]
    expected_test = [2016, 2017, 2018]

    assert np.array_equal(x_train.anchor_year, expected_train)
    xr.testing.assert_equal(x_test, x1.sel(anchor_year=expected_test))
    assert np.array_equal(y_train.anchor_year, expected_train)
    xr.testing.assert_equal(y_test, y.sel(anchor_year=expected_test))


def test_kfold_xxy(dummy_data):
    """Correctly split x1, x2, and y."""
    x1, x2, y = dummy_data
    cv = lilio.traintest.TrainTestSplit(KFold(n_splits=3))
    x_train, x_test, y_train, y_test = next(cv.split(x1, x2, y=y))
    expected_train = [2019, 2020, 2021, 2022]
    expected_test = [2016, 2017, 2018]

    assert np.array_equal(x_train[0].anchor_year, expected_train)
    xr.testing.assert_equal(x_test[1], x2.sel(anchor_year=expected_test))
    assert np.array_equal(y_train.anchor_year, expected_train)
    xr.testing.assert_equal(y_test, y.sel(anchor_year=expected_test))


def test_kfold_too_short(dummy_data):
    "Fail if there is only a single anchor year: no splits can be made"
    x1, _, _ = dummy_data
    x = x1.isel(anchor_year=1)
    cv = lilio.traintest.TrainTestSplit(KFold(n_splits=3))

    with pytest.raises(ValueError):
        next(cv.split(x))


def test_kfold_different_xcoords(dummy_data):
    x1, x2, _ = dummy_data
    x1 = x1.isel(anchor_year=slice(1, None, None))
    cv = lilio.traintest.TrainTestSplit(KFold(n_splits=3))

    with pytest.raises(lilio.traintest.CoordinateMismatchError):
        next(cv.split(x1, x2))


def test_custom_dim(dummy_data):
    x1, _, _ = dummy_data
    x = x1.rename(anchor_year="custom_coord")
    cv = lilio.traintest.TrainTestSplit(KFold(n_splits=3))
    x_train, _ = next(cv.split(x, dim="custom_coord"))
    expected_train = [2019, 2020, 2021, 2022]

    assert np.array_equal(x_train.custom_coord, expected_train)



"""Lilio's main Calendar module."""
import copy
import re
import warnings
from os import linesep
from typing import List
from typing import Literal
from typing import Tuple
from typing import Union
import pandas as pd
import xarray as xr
from pandas.tseries.offsets import DateOffset
from . import _plot
from . import utils


_MappingYears = Tuple[Literal["years"], int, int]
_MappingData = Tuple[Literal["data"], pd.Timestamp, pd.Timestamp]


class Interval:
    """Basic construction element of calendar for defining precursors and targets."""

    def __init__(
        self,
        role: Literal["target", "precursor"],
        length: Union[str, dict],
        gap: Union[str, dict] = "0d",
    ) -> None:
        """Construct the basic element of the calendar.

        The Interval is characterised by its type (either target or precursor), its
        length and the gap between it and the previous interval of its type (or the
        anchor date, if the interval is the first target/first precursor).

        Args:
            role: The type of interval. Either "target" or "precursor".
            length: The length of the interval. This can either be a pandas-like
                frequency string (e.g. "10d", "2W", or "3M"), or a pandas.DateOffset
                compatible dictionary such as {days=10}, {weeks=2}, or
                {months=1, weeks=2}.
            gap: The gap between the previous interval and this interval. Valid inputs
                are the same as the length keyword argument. Defaults to "0d".

        Example:
            >>> from lilio import Interval
            >>> iv = Interval("target", length="7d")
            >>> iv
            Interval(role='target', length='7d', gap='0d')

            You can modify the interval's properties in-place:

            >>> iv.gap = "1W"
            >>> iv
            Interval(role='target', length='7d', gap='1W')

        """
        self.length = length
        self.gap = gap
        self._role = role
        self._target = role == "target"

        self._gap_dateoffset: pd.DateOffset
        self._length_dateoffset: pd.DateOffset

        # TO DO: support lead_time
        # self.lead_time = lead_time

    @property
    def is_target(self):
        """Return whether this Intervals is a target interval."""
        return self._target

    @property
    def role(self):
        """Return the type of interval."""
        return self._role

    @property
    def length(self):
        """Return the length of the interval, as a pandas.DateOffset."""
        return self._length

    @length.setter
    def length(self, value: Union[str, dict]):
        self._length = value
        if isinstance(value, str):
            self._length_dateoffset = DateOffset(
                **utils.parse_freqstr_to_dateoffset(value)
            )
        else:
            self._length_dateoffset = DateOffset(**value)

    @property
    def length_dateoffset(self):
        """Return the length property as a dateoffset."""
        return self._length_dateoffset

    @property
    def gap(self):
        """Returns the gap of the interval, as a pandas.DateOffset."""
        return self._gap

    @gap.setter
    def gap(self, value: Union[str, dict]):
        self._gap = value
        if isinstance(value, str):
            self._gap_dateoffset = DateOffset(
                **utils.parse_freqstr_to_dateoffset(value)
            )
        else:
            self._gap_dateoffset = DateOffset(**value)

    @property
    def gap_dateoffset(self):
        """Get the gap property as a dateoffset."""
        return self._gap_dateoffset

    def __repr__(self):
        """Return a string representation of the Interval class."""
        props = [
            ("role", self.role),
            ("length", self.length),
            ("gap", self.gap),
        ]

        propstr = ", ".join([f"{k}={repr(v)}" for k, v in props])
        return f"{self.__class__.__name__}({propstr})"


class Calendar:
    """Build a calendar from scratch with basic construction elements."""

    def __init__(
        self,
        anchor: str,
        allow_overlap: bool = False,
        mapping: Union[
            None,
            _MappingYears,
            _MappingData,
        ] = None,
        intervals: Union[None, List[Interval]] = None,
    ):
        """Instantiate a basic container for building calendar using basic blocks.

        This is a highly flexible calendar which allows the user to build their own
        calendar with the basic building blocks of target and precursor periods.

        Users have the freedom to create calendar with customized intervals, gap
        between intervals, and even overlapped intervals. They need to manage the
        calendar themselves.

        Some shorthand calendars, such as a `daily_calendar`, `weekly_calendar` and
        `monthly_calendar` are available in lilio.calendar_shorthands. These can be used
        to easily construct basic calendars with only a few parameters, but do not have
        the flexibility that this calendar builder module provides.

        Args:
            anchor: String denoting the anchor date. The following inputs are valid:
                    - "MM-DD" for a month and day. E.g. "12-31".
                    - "MM" for only a month, e.g. "4" for March.
                    - English names and abbreviations of months. E.g. "December" or "jan".
                    - "Www" for a week number, e.g. "W05" for the fifth week of the year.
                    - "Www-D" for a week number plus day of week. E.g. "W01-4" for the
                        first thursday of the year.
            allow_overlap: If overlapping intervals between years is allowed or not.
                Default behaviour is False, which means that anchor years will be
                skipped to avoid data being shared between anchor years.
            mapping: Calendar mapping. Input in the form: ("years", 2000, 2020) or
                ("data", pd.Timestamp("2000-01-01"), pd.Timestamp("2020-01-01")). The
                calendar mapping is usually set with the `map_years` or `map_to_data`
                methods.
            intervals: A list of Interval objects that should be appended to the
                calendar when it is initialized.

        Example:
            Instantiate a custom calendar and appending target/precursor periods.

            >>> import lilio
            >>> calendar = lilio.Calendar(anchor="12-31")
            >>> calendar  # doctest: +NORMALIZE_WHITESPACE
            Calendar(
                anchor='12-31',
                allow_overlap=False,
                mapping=None,
                intervals=None
            )

        """
        self._anchor, self._anchor_fmt = self._parse_anchor(anchor)
        self._allow_overlap = allow_overlap
        self.targets: List[Interval] = []
        self.precursors: List[Interval] = []

        self._first_year: Union[None, int] = None
        self._last_year: Union[None, int] = None

        if intervals is not None:
            # pylint: disable=expression-not-assigned
            [self._append(iv) for iv in intervals]

        self._mapping: Union[None, Literal["years", "data"]]
        self._set_mapping(mapping)

    @property
    def n_targets(self):
        """Return the number of targets."""
        return len(self.targets)

    @property
    def anchor(self):
        """Return the anchor."""
        return self._anchor

    @anchor.setter
    def anchor(self, value):
        self._anchor, self._anchor_fmt = self._parse_anchor(value)

    @property
    def allow_overlap(self):
        """Returns the allow_overlap: if overlapping intervals are allowed or not."""
        return self._allow_overlap

    @allow_overlap.setter
    def allow_overlap(self, value: bool):
        if isinstance(value, bool):
            self._allow_overlap = value
        else:
            raise ValueError(
                f"allow_overlap should be either True or False, not {value}"
                f"of type {type(value)}"
            )

    @property
    def mapping(self) -> Union[None, Literal["years", "data"]]:
        """Return the mapping of the calendar. Either  None, "years", or "data"."""
        return self._mapping

    def add_intervals(
        self,
        role: Literal["target", "precursor"],
        length: str,
        gap: str = "0d",
        n: int = 1,
    ) -> None:
        """Add one or more intervals to the calendar.

        The interval can be a target or a precursor, and can be defined by its length,
        a possible gap between this interval and the preceding interval.

        Args:
            role: Either a 'target' or 'precursor' interval(s).
            length: The length of the interval(s), in a format of '5d' for five days, '2W'
                for two weeks, or '1M' for one month.
            gap: The gap between this interval and the preceding target/precursor
                interval. Same format as the length argument.
            n: The number of intervals which should be added to the calendar. Defaults
                to 1.
        """
        if not isinstance(n, int):
            raise ValueError(
                "Please input an 'int' type for the 'n' argument." f" Not a {type(n)}."
            )
        if n <= 0:
            raise ValueError(
                "The number of intervals 'n' has to be 1 or greater, " f"not '{n}'."
            )

        if role in ["target", "precursor"]:
            for _ in range(n):
                self._append(Interval(role, length, gap))
        else:
            raise ValueError(
                f"Type '{role}' is not a valid interval type. Please "
                "choose between 'target' and 'precursor'"
            )

    def _get_anchor(self, year: int) -> pd.Timestamp:
        """Generate an anchor timestamp for your specific calendar.

        The method should return the exact timestamp of the end of the anchor_year's
        0 interval, e.g., for the AdventCalendar:
        pd.Timestamp(year, self.month, self.day)

        Args:
            year (int): anchor year for which the anchor timestamp should be generated

        Returns:
            pd.Timestamp: Timestamp at the end of the anchor_years interval 0.
        """
        return pd.to_datetime(
            f"{year}-" + self._anchor, format="%Y-" + self._anchor_fmt
        )

    def _parse_anchor(self, anchor_str: str) -> Tuple[str, str]:
        """Parse the user-input anchor.

        Args:
            anchor_str: Anchor string in the right formatting.

        Returns:
            Datetime formatter to parse the anchor into a date.
        """
        # non string check
        if not isinstance(anchor_str, str):
            raise ValueError("Anchor input must be a string with expected format.")
        # format match
        if re.fullmatch("\\d{1,2}-\\d{1,2}", anchor_str):
            utils.check_month_day(*[int(x) for x in anchor_str.split("-")])
            fmt = "%m-%d"
        elif re.fullmatch("\\d{1,2}", anchor_str):
            utils.check_month_day(int(anchor_str))
            fmt = "%m"
        elif re.fullmatch("W\\d{1,2}-\\d", anchor_str):
            utils.check_week_day(*[int(x) for x in anchor_str[1:].split("-")])
            fmt = "W%W-%w"
        elif re.fullmatch("W\\d{1,2}", anchor_str):
            utils.check_week_day(int(anchor_str[1:]))
            fmt = "W%W-%w"
            anchor_str += "-1"
        elif anchor_str.lower() in utils.get_month_names():
            anchor_str = str(utils.get_month_names()[anchor_str.lower()])
            fmt = "%m"
        else:
            raise ValueError(
                f"Anchor input '{anchor_str}' does not match expected format."
            )
        return anchor_str, fmt

    def _append(self, interval):
        """Append target/precursor periods to the calendar."""
        if interval.is_target:
            self.targets.append(interval)
        else:
            self.precursors.append(interval)

    def _map_year(self, year: int) -> pd.Series:
        """Return a concrete IntervalIndex for the given year.

        Since our calendars are used to study periodic events, they are first
        instantiated without specific year(s). This method adds a specific year
        to the calendar and returns an intervalindex, mapping the
        Calendar to the given year.

        Intended for internal use, in conjunction with map_years or map_to_data.

        Args:
            year: The year for which the Calendar will be realized

        Returns:
            Pandas Series filled with Intervals of the calendar's frequency, counting
            backwards from the calendar's achor.
        """
        intervals_target = self._concatenate_periods(year, self.targets, True)
        intervals_precursor = self._concatenate_periods(year, self.precursors, False)

        year_intervals = intervals_precursor[::-1] + intervals_target

        # turn the list of intervals into pandas series
        year_intervals = pd.Series(year_intervals[::-1], name=year)
        year_intervals.index.name = "i_interval"
        return year_intervals

    def _concatenate_periods(self, year, list_periods, is_target):
        # generate intervals based on the building blocks
        intervals = []
        if is_target:
            # build from left to right
            left_date = self._get_anchor(year)
            # loop through all the building blocks to
            for block in list_periods:
                left_date += block.gap_dateoffset
                right_date = left_date + block.length_dateoffset
                intervals.append(pd.Interval(left_date, right_date, closed="left"))
                # update left date
                left_date = right_date
        else:
            # build from right to left
            right_date = self._get_anchor(year)
            # loop through all the building blocks to
            for block in list_periods:
                right_date -= block.gap_dateoffset
                left_date = right_date - block.length_dateoffset
                intervals.append(pd.Interval(left_date, right_date, closed="left"))
                # update right date
                right_date = left_date

        return intervals

    def _get_skip_nyears(self) -> int:
        """Determine how many years need to be skipped to avoid overlapping data.

        Required to prevent information leakage between anchor years.

        Returns:
            int: Number of years that need to be skipped.
        """
        if self._allow_overlap:
            return 0

        proto_year = 2000
        skip_years = 0

        start_calendar = self._get_anchor(proto_year)
        for prec in self.precursors:
            start_calendar -= prec.gap_dateoffset
            start_calendar -= prec.length_dateoffset

        while True:
            prev_end_calendar = self._get_anchor(proto_year - 1 - skip_years)
            for target in self.targets:
                prev_end_calendar += target.gap_dateoffset
                prev_end_calendar += target.length_dateoffset
            if prev_end_calendar > start_calendar:
                skip_years += 1
            else:
                break

        return skip_years

    def map_years(self, start: int, end: int):
        """Add a start and end year mapping to the calendar.

        If the start and end years are the same, the intervals for only that single
        year are returned by calendar.get_intervals().

        Args:
            start: The first year for which the calendar will be realized
            end: The last year for which the calendar will be realized

        Returns:
            The calendar mapped to the input start and end year.
        """
        if start > end:
            raise ValueError("The start year cannot be greater than the end year")

        self._first_year = start
        self._last_year = end
        self._mapping = "years"

        self._first_timestamp = None
        self._last_timestamp = None

        return self

    def map_to_data(
        self,
        input_data: Union[pd.Series, pd.DataFrame, xr.Dataset, xr.DataArray],
    ):
        """Map the calendar to input data period.

        Stores the first and last intervals of the input data to the calendar, so that
        the intervals can cover the data to the greatest extent.

        Args:
            input_data: Input data for datetime mapping. Its index must be either
                pandas.DatetimeIndex, or an xarray `time` coordinate with datetime
                data.

        Returns:
            The calendar mapped to the input data period.
        """
        utils.check_timeseries(input_data)

        # check the datetime order of input data
        if isinstance(input_data, (pd.Series, pd.DataFrame)):
            self._first_timestamp = input_data.index.min()
            self._last_timestamp = input_data.index.max()
        else:
            self._first_timestamp = pd.Timestamp(input_data.time.min().values)
            self._last_timestamp = pd.Timestamp(input_data.time.max().values)

        self._mapping = "data"
        self._first_year = None
        self._last_year = None

        return self

    def _set_year_range_from_timestamps(self):
        min_year = self._first_timestamp.year  # type: ignore
        max_year = self._last_timestamp.year  # type: ignore

        # ensure that the input data could always cover the advent calendar
        # last date check
        if self._map_year(max_year).iloc[0].right > self._last_timestamp:
            max_year -= 1
        # first date check
        while self._map_year(min_year).iloc[-1].right <= self._first_timestamp:
            min_year += 1

        # map year(s) and generate year realized advent calendar
        if max_year >= min_year:
            self._first_year = min_year
            self._last_year = max_year
        else:
            raise ValueError(
                "The input data could not cover the target advent calendar."
            )

        return self

    def _set_mapping(self, mapping):
        if mapping is None:
            self._mapping = None
        elif mapping[0] == "years":
            self.map_years(mapping[1], mapping[2])
        elif mapping[0] == "data":
            self._mapping = "data"
            self._first_timestamp = mapping[1]
            self._last_timestamp = mapping[2]
        else:
            raise ValueError(
                "Unknown mapping passed to calendar. Valid options are"
                "either 'years' or 'data'."
            )

    def _rename_intervals(self, intervals: pd.DataFrame) -> pd.DataFrame:
        """Add target labels to the header row of the intervals.

        Args:
            intervals (pd.Dataframe): Dataframe with intervals.

        Returns:
            pd.Dataframe: Dataframe with target periods labelled, sorted by i_interval value.
        """
        # rename precursors
        intervals = intervals.rename(
            columns={
                i: self.n_targets - i - 1
                for i in range(self.n_targets, len(intervals.columns))
            }
        )

        # rename targets
        intervals = intervals.rename(
            columns={i: self.n_targets - i for i in range(self.n_targets)}
        )

        return intervals.sort_index(axis=1)

    def get_intervals(self) -> pd.DataFrame:
        """Retrieve updated intervals from the Calendar object."""
        if self._mapping is None:
            raise ValueError(
                "Cannot retrieve intervals without map_years or "
                "map_to_data having configured the calendar."
            )
        if self._mapping == "data":
            self._set_year_range_from_timestamps()

        year_range = range(
            self._last_year, self._first_year - 1, -(self._get_skip_nyears() + 1)  # type: ignore
        )

        intervals = pd.concat([self._map_year(year) for year in year_range], axis=1).T

        intervals = self._rename_intervals(intervals)

        intervals.index.name = "anchor_year"
        return intervals.sort_index(axis=0, ascending=False)

    def show(self) -> pd.DataFrame:
        """Display the intervals the Calendar will generate for the current setup.

        Returns:
            pd.Dataframe: Dataframe containing the calendar intervals.
        """
        return self.get_intervals()

    def __repr__(self) -> str:
        """Return a string representation of the Calendar."""
        intervals = self.targets + self.precursors
        if len(intervals) == 0:
            intervals_str = repr(None)
        else:
            intervals_str = (
                f"[{linesep}\t\t"
                + f",{linesep}\t\t".join([repr(iv) for iv in intervals])
                + f"{linesep}\t]"
            )

        if self._mapping == "years":
            mapping = ("years", self._first_year, self._last_year)
        elif self._mapping == "data":
            mapping = ("data", self._first_timestamp, self._last_timestamp)
        else:
            mapping = None

        props = [
            ("anchor", repr(self.anchor)),
            ("allow_overlap", repr(self.allow_overlap)),
            ("mapping", repr(mapping)),
            ("intervals", intervals_str),
        ]

        propstr = f"{linesep}\t" + f",{linesep}\t".join([f"{k}={v}" for k, v in props])
        return f"{self.__class__.__name__}({propstr}{linesep})".replace("\t", "    ")

    def visualize(  # noqa: PLR0913 (too-many-arguments)
        self,
        n_years: int = 3,
        interactive: bool = False,
        relative_dates: bool = False,
        show_length: bool = False,
        add_legend: bool = True,
        ax=None,
        **bokeh_kwargs,
    ) -> None:
        """Plot a visualization of the current calendar setup, to aid in user setup.

        Note: The interactive visualization requires the `bokeh` package to be installed
        in the active Python environment.

        Args:
            n_years: Sets the maximum number of anchor years that should be shown. By
                default only the most recent 3 are visualized, to ensure that they
                fit within the plot.
            interactive: If False, matplotlib will be used for the visualization. If
                True, bokeh will be used.
            relative_dates: Toggles if the intervals should be displayed relative to the
                anchor date, or as absolute dates.
            show_length: Toggles if the frequency of the intervals should be displayed.
                Defaults to False (Matplotlib plotter only).
            add_legend: Toggles if a legend should be added to the plot (Matplotlib only)
            ax: Matplotlib axis object to plot the visualization into.
            **bokeh_kwargs: Keyword arguments to pass to Bokeh's plotting.Figure. See
                https://docs.bokeh.org/en/2.4.2/docs/reference/plotting/figure.html
                for a list of possible keyword arguments.
        """
        calendar = copy.deepcopy(self)
        if calendar._mapping is None:  # pylint: disable=protected-access
            calendar.map_years(2000, 2000)
            if not relative_dates:
                print(
                    "Setting relative_dates=True. To see absolute dates, first call "
                    "calendar.map_years or calendar.map_data"
                )
                relative_dates = True
            add_yticklabels = False
        else:
            add_yticklabels = True

        n_years = max(n_years, 1)
        n_years = min(n_years, len(calendar.get_intervals().index))

        if interactive:
            utils.assert_bokeh_available()
            # pylint: disable=import-outside-toplevel
            from ._bokeh_plots import bokeh_visualization

            if ax is not None:
                warnings.warn(
                    "ax is only a valid keyword argument for the non-interactive "
                    "matplotlib backend. Bokeh's figure can be controlled by passing "
                    "Bokeh Figure keyword arguments (e.g. width=800).",
                    UserWarning,
                )
            bokeh_visualization(
                calendar, n_years, relative_dates, add_yticklabels, **bokeh_kwargs
            )
        else:
            if bokeh_kwargs:
                warnings.warn(
                    "kwargs for bokeh have been passed to visualize(), but the "
                    "matplotlib backend does not support these. Use the 'ax' kwarg "
                    "instead to control the generated figure.",
                    UserWarning,
                )

            _plot.matplotlib_visualization(
                calendar,
                n_years,
                relative_dates,
                show_length,
                add_legend,
                add_yticklabels,
                ax=ax,
            )

    @property
    def flat(self) -> pd.DataFrame:
        """Returns the flattened intervals."""
        return self.get_intervals().stack()  # type: ignore



"""Commonly used utility functions for Lilio."""
import re
import typing
import warnings
from typing import Dict
from typing import Union
import numpy as np
import pandas as pd
import xarray as xr


if typing.TYPE_CHECKING:
    from lilio import Calendar


MONTH_LENGTH = 30  # Month length for Timedelta checks.


def check_timeseries(
    data: Union[pd.Series, pd.DataFrame, xr.DataArray, xr.Dataset]
) -> None:
    """Check if input data contains valid time data.

    Checks if:
     - Input data is pd.Dataframe/pd.Series/xr.Dataset/xr.DataArray.
     - Input data has a time index (pd), or a dim named `time` containing datetime
       values
    """
    if not isinstance(data, (pd.Series, pd.DataFrame, xr.DataArray, xr.Dataset)):
        raise ValueError("The input data is neither a pandas or xarray object")
    if isinstance(data, (pd.Series, pd.DataFrame)):
        check_time_dim_pandas(data)
    elif isinstance(data, (xr.DataArray, xr.Dataset)):
        check_time_dim_xarray(data)


def check_time_dim_xarray(data) -> None:
    """Check if an xarray data has a time dimensions with time data."""
    if "time" not in data.dims:
        raise ValueError(
            "The input DataArray/Dataset does not contain a `time` dimension"
        )
    if not xr.core.common.is_np_datetime_like(data["time"].dtype):  # type: ignore
        raise ValueError("The `time` dimension is not of a datetime format")


def check_time_dim_pandas(data) -> None:
    """Check if pandas data has an index with time data."""
    if not isinstance(data.index, pd.DatetimeIndex):
        raise ValueError("The input data does not have a datetime index.")


def check_empty_intervals(data: Union[pd.DataFrame, xr.Dataset]) -> None:
    """Check for empty intervals in pandas data.

    Note: For the Dataset, all values within a certain interval, anchor_year combination
    have to be NaN, to allow for, e.g., empty gridcells in a latitude/longitude grid.

    Args:
        data (Union[pd.DataFrame, xr.Dataset]): Data that should be checked for empty
            intervals. Should be done after resampling the data.

    Raises:
        UserWarning: If the data is insufficient.

    Returns:
        None
    """
    if isinstance(data, pd.DataFrame) and not np.any(np.isnan(data.iloc[:, 3:])):
        return None
    if isinstance(data, xr.Dataset):
        time_vars = [var for var in data.data_vars if "anchor_year" in data[var].dims]
        if not any(
            data[var].isnull().any(dim=["i_interval", "anchor_year"]).all()
            for var in time_vars
        ):
            return None

    warnings.warn(
        "The input data could not fully cover the calendar's intervals. "
        "Intervals without available data will contain NaN values."
    )
    return None


def infer_input_data_freq(
    data: Union[pd.Series, pd.DataFrame, xr.DataArray, xr.Dataset]
) -> pd.Timedelta:
    """Infer the frequency of the input data, for comparison with the calendar freq.

    Args:
        data: Pandas or xarray data to infer the frequency of.

    Returns:
        a pd.Timedelta
    """
    if isinstance(data, (pd.Series, pd.DataFrame)):
        data_freq = pd.infer_freq(data.index)
        if data_freq is None:  # Manually infer the frequency
            data_freq = np.min(data.index.values[1:] - data.index.values[:-1])
    else:
        data_freq = xr.infer_freq(data.time)
        if data_freq is None:  # Manually infer the frequency
            data_freq = (data.time.values[1:] - data.time.values[:-1]).min()

    if isinstance(data_freq, str):
        data_freq.replace("-", "")  # Get the absolute frequency

        if not re.match(r"\d+\D", data_freq):  # infer_freq can return "d" for "1d".
            data_freq = "1" + data_freq

        data_freq = (  # Deal with monthly timedelta case
            replace_month_length(data_freq) if data_freq[-1] == "M" else data_freq
        )
    return pd.Timedelta(data_freq)


def replace_month_length(length: str) -> str:
    """Replace month lengths with an equivalent length in days."""
    ndays = int(length[:-1]) * MONTH_LENGTH
    return f"{ndays}d"


def get_smallest_calendar_freq(calendar: "Calendar") -> pd.Timedelta:
    """Return the smallest length of the calendar's intervals as a Timedelta."""
    intervals = calendar.targets + calendar.precursors
    lengthstr = [iv.length for iv in intervals]
    lengthstr = [ln.replace("-", "") for ln in lengthstr]  # Account for neg. lengths
    lengthstr = [replace_month_length(ln) if ln[-1] == "M" else ln for ln in lengthstr]
    lengths = [pd.Timedelta(ln) for ln in lengthstr]
    return min(lengths)


def check_input_frequency(
    calendar: "Calendar", data: Union[pd.Series, pd.DataFrame, xr.DataArray, xr.Dataset]
):
    """Compare the frequency of (input) data to the frequency of the calendar.

    Note: Pandas and xarray have the builtin function `infer_freq`, but this function is
    not robust enough for our purpose, so we have to manually infer the frequency if the
    builtin one fails.
    """
    data_freq = infer_input_data_freq(data)
    calendar_freq = get_smallest_calendar_freq(calendar)

    if calendar_freq < data_freq:
        raise ValueError(
            "The data is of a lower time resolution than the calendar. "
            "This would lead to incorrect data and/or NaN values in the resampled data."
            " Please make the Calendar's intervals larger, or use data of a higher time"
            " resolution."
            f"\nInfered data frequency: {str(data_freq)} < calendar frequency "
            f"{str(calendar_freq)}"
        )
    if calendar_freq < 2 * data_freq:
        warnings.warn(
            "The input data frequency is very close to the Calendar's frequency. "
            "This could lead to issues like aliasing or incorrect resampling. "
            "If possible: make the Calendar's intervals larger, or use data of a higher"
            " time resolution."
        )


def convert_interval_to_bounds(data: xr.Dataset) -> xr.Dataset:
    """Convert pandas intervals to bounds in a xarray Dataset.

    pd.Interval objects cannot be written to netCDF. To allow writing the
    calendar-resampled data to netCDF these intervals have to be converted to bounds.
    This function adds a 'bounds' dimension, with 'left' and 'right' coordinates, and
    converts the 'interval' coordinates to this system.

    Args:
        data: Input data with intervals as pd.Interval objects.

    Returns:
        Input data with the intervals converted to bounds.
    """
    data = data.stack(coord=["anchor_year", "i_interval"])
    bounds = np.array([[val.left, val.right] for val in data.interval.values])
    data["left_bound"] = ("coord", bounds[:, 0])
    data["right_bound"] = ("coord", bounds[:, 1])
    data["left_bound"].attrs = {
        "name": "Left bound of the interval",
        "closed": "True",
    }
    data["right_bound"].attrs = {
        "name": "Right bound of the interval",
        "closed": "False",
    }
    data = data.unstack("coord")
    data = data.drop_vars(["interval"])
    data = data.set_coords(["left_bound", "right_bound"])
    return data


def check_reserved_names(
    input_data: Union[pd.Series, pd.DataFrame, xr.DataArray, xr.Dataset]
) -> None:
    """Check if reserved names are already in the input data. E.g. "anchor_year"."""
    reserved_names_pd = ["anchor_year", "i_interval", "is_target"]
    reserved_names_xr = reserved_names_pd + ["left_bound", "right_bound"]

    if isinstance(input_data, pd.DataFrame):
        if any(name in input_data.columns for name in reserved_names_pd):
            raise ValueError(
                "The input data contains one or more reserved names. Please remove or "
                f"rename these before continuing.\n Reserved names: {reserved_names_pd}"
            )
    elif isinstance(input_data, (xr.DataArray, xr.Dataset)):
        data_names = [
            input_data.keys()
            if isinstance(input_data, xr.Dataset)
            else list(input_data.coords) + [input_data.name]
        ]
        if any(name in data_names for name in reserved_names_xr):
            raise ValueError(
                "The input data contains one or more reserved names. Please remove or "
                f"rename these before continuing.\n Reserved names: {reserved_names_xr}"
            )


def assert_bokeh_available():
    """Util that attempts to load the optional module bokeh."""
    try:
        import bokeh as _  # noqa: F401 (unused import)

    except ImportError as e:
        raise ImportError(
            "Could not import the `bokeh` module.\nPlease install this"
            " before continuing, with either `pip` or `conda`."
        ) from e


def get_month_names() -> Dict:
    """Generate a dictionary with English lowercase month names and abbreviations.

    Returns:
        Dictionary containing the English names of the months, including their
            abbreviations, linked to the number of each month.
            E.g. {'december': 12, 'jan': 1}
    """
    return {
        "january": 1,
        "february": 2,
        "march": 3,
        "april": 4,
        "may": 5,
        "june": 6,
        "july": 7,
        "august": 8,
        "september": 9,
        "october": 10,
        "november": 11,
        "december": 12,
        "jan": 1,
        "feb": 2,
        "mar": 3,
        "apr": 4,
        "jun": 6,
        "jul": 7,
        "aug": 8,
        "sep": 9,
        "oct": 10,
        "nov": 11,
        "dec": 12,
    }


def check_month_day(month: int, day: int = 1):
    """Check if the input day/month combination is valid.

    Months must be between 1 and 12, and days must be within 1 and 28/30/31 (depending
    on the month).

    Args:
        month: Month number
        day: Day number. Defaults to 1.
    """
    if month in {1, 3, 5, 7, 8, 10, 12}:
        if (day < 1) or (day > 31):
            raise ValueError(
                "Incorrect anchor input. "
                f"Day number {day} is not a valid day for month {month}"
            )
    elif month in {4, 6, 9, 11}:
        if (day < 1) or (day > 30):
            raise ValueError(
                "Incorrect anchor input. "
                f"Day number {day} is not a valid day for month {month}"
            )
    elif month == 2:
        if (day < 1) or (day > 28):
            raise ValueError(
                "Incorrect anchor input. "
                f"Day number {day} is not a valid day for month {month}"
            )
    else:
        raise ValueError(
            "Incorrect anchor input. Month number must be between 1 and 12."
        )


def check_week_day(week: int, day: int = 1):
    """Validate the week and day numbers."""
    if week == 53:
        raise ValueError(
            "Incorrect anchor input. "
            "Week 53 is not a valid input, as not every year contains a 53rd week."
        )
    if (week < 1) or (week > 52):
        raise ValueError(
            "Incorrect anchor input. Week numbers must be between 1 and 52."
        )
    if (day < 1) or (day > 7):
        raise ValueError(
            "Incorrect anchor input. Weekday numbers must be between 1 and 7."
        )


def parse_freqstr_to_dateoffset(time_str):
    """Parse the user-input time strings.

    Args:
        time_str: Time length string in the right formatting.

    Returns:
        Dictionary as keyword argument for Pandas DateOffset.
    """
    if re.fullmatch(r"[+-]?\d*d", time_str):
        time_dict = {"days": int(time_str[:-1])}
    elif re.fullmatch(r"[+-]?\d*M", time_str):
        time_dict = {"months": int(time_str[:-1])}
    elif re.fullmatch(r"[+-]?\d*W", time_str):
        time_dict = {"weeks": int(time_str[:-1])}
    else:
        raise ValueError("Please input a time string in the correct format.")

    return time_dict



"""Tests for the lilio.Calendar module.
"""
from typing import Literal
import numpy as np
import pandas as pd
import pytest
import xarray as xr
from pandas.tseries.offsets import DateOffset
from lilio import Calendar
from lilio import Interval
from lilio import daily_calendar


def interval(start, end, closed: Literal["left", "right", "both", "neither"] = "left"):
    """Shorthand for more readable tests."""
    return pd.Interval(pd.Timestamp(start), pd.Timestamp(end), closed=closed)


class TestInterval:
    """Test the Interval class."""

    def test_target_interval(self):
        target = Interval("target", "20d", "10d")
        assert isinstance(target, Interval)
        assert target.length_dateoffset == DateOffset(days=20)
        assert target.gap_dateoffset == DateOffset(days=10)
        assert target.is_target

    def test_precursor_interval(self):
        precursor = Interval("precursor", "20d", "10d")
        assert isinstance(precursor, Interval)
        assert precursor.length_dateoffset == DateOffset(days=20)
        assert precursor.gap_dateoffset == DateOffset(days=10)
        assert not precursor.is_target

    def test_interval_months(self):
        target = Interval("target", "2M", "1M")
        assert target.length_dateoffset == DateOffset(months=2)
        assert target.gap_dateoffset == DateOffset(months=1)

    def test_interval_weeks(self):
        target = Interval("target", "3W", "2W")
        assert target.length_dateoffset == DateOffset(weeks=3)
        assert target.gap_dateoffset == DateOffset(weeks=2)

    def test_target_interval_dict(self):
        a = {"months": 1, "weeks": 2, "days": 1}
        b = {"months": 2, "weeks": 1, "days": 5}
        target = Interval("target", length=a, gap=b)
        assert target.length_dateoffset == DateOffset(**a)
        assert target.gap_dateoffset == DateOffset(**b)

    def test_repr(self):
        target = Interval("target", "20d", "10d")
        expected = "Interval(role='target', length='20d', gap='10d')"
        assert repr(target) == expected

    def test_repr_eval(self):
        target = Interval("target", "20d", "10d")
        _ = eval(repr(target))  # pylint: disable=eval-used


class TestCalendar:
    """Test the (custom) Calendar methods."""

    @pytest.fixture
    def dummy_calendar(self):
        cal = Calendar(anchor="12-31")
        # append building blocks
        cal.add_intervals("target", "20d")
        cal.add_intervals("precursor", "10d")
        # map years
        cal = cal.map_years(2021, 2021)
        return cal

    def test_init(self):
        cal = Calendar(anchor="12-31")
        assert isinstance(cal, Calendar)

    def test_repr_basic(self):
        cal = Calendar(anchor="12-31")
        expected = (
            "Calendar(anchor='12-31',allow_overlap=False,mapping=None,intervals=None)"
        )
        calrepr = repr(cal)

        # Test that the repr can be pasted back into the terminal
        _ = eval(calrepr)  # pylint: disable=eval-used

        # remove whitespaces:
        calrepr = calrepr.replace(" ", "").replace("\r", "").replace("\n", "")
        assert calrepr == expected

    def test_repr_reproducible(self):
        cal = Calendar(anchor="12-31", allow_overlap=True)
        cal.add_intervals("target", "10d")
        cal.map_years(2020, 2022)
        repr_dict = eval(repr(cal)).__dict__  # pylint: disable=eval-used
        assert repr_dict["_anchor"] == "12-31"
        assert repr_dict["_mapping"] == "years"
        assert repr_dict["_first_year"] == 2020
        assert repr_dict["_last_year"] == 2022
        assert repr_dict["_allow_overlap"] is True
        assert (
            repr(repr_dict["targets"][0])
            == "Interval(role='target', length='10d', gap='0d')"
        )

    def test_show(self, dummy_calendar):
        expected_calendar_repr = (
            "i_interval -1 1\n anchor_year \n 2021"
            + "[2021-12-21, 2021-12-31) [2021-12-31, 2022-01-20)"
        )
        expected_calendar_repr = expected_calendar_repr.replace(" ", "")
        assert repr(dummy_calendar.show()).replace(" ", "") == expected_calendar_repr

    def test_no_intervals(self):
        cal = Calendar(anchor="12-31")
        with pytest.raises(ValueError):
            cal.get_intervals()

    def test_flat(self, dummy_calendar):
        expected = np.array(
            [
                interval("2021-12-21", "2021-12-31", closed="left"),
                interval("2021-12-31", "2022-01-20", closed="left"),
            ]
        )
        assert np.array_equal(dummy_calendar.flat, expected)

    def test_add_intervals(self, dummy_calendar):
        dummy_calendar.add_intervals("target", "30d")
        dummy_calendar = dummy_calendar.map_years(2021, 2021)
        expected = np.array(
            [
                interval("2021-12-21", "2021-12-31", closed="left"),
                interval("2021-12-31", "2022-01-20", closed="left"),
                interval("2022-01-20", "2022-02-19", closed="left"),
            ]
        )
        assert np.array_equal(dummy_calendar.flat, expected)

    def test_add_intervals_multiple(self, dummy_calendar):
        dummy_calendar.add_intervals("target", "30d", n=2)
        dummy_calendar = dummy_calendar.map_years(2021, 2021)
        expected = np.array(
            [
                interval("2021-12-21", "2021-12-31", closed="left"),
                interval("2021-12-31", "2022-01-20", closed="left"),
                interval("2022-01-20", "2022-02-19", closed="left"),
                interval("2022-02-19", "2022-03-21", closed="left"),
            ]
        )
        assert np.array_equal(dummy_calendar.flat, expected)

    @pytest.mark.parametrize("incorrect_n", (2.0, [1], 0, -10))  # non-int or <=0.
    def test_add_intervals_incorrect_n(self, dummy_calendar, incorrect_n):
        with pytest.raises(ValueError):
            dummy_calendar.add_intervals("target", "30d", n=incorrect_n)

    def test_gap_intervals(self, dummy_calendar):
        dummy_calendar.add_intervals("target", "20d", gap="10d")
        dummy_calendar = dummy_calendar.map_years(2021, 2021)
        expected = np.array(
            [
                interval("2021-12-21", "2021-12-31", closed="left"),
                interval("2021-12-31", "2022-01-20", closed="left"),
                interval("2022-01-30", "2022-02-19", closed="left"),
            ]
        )
        assert np.array_equal(dummy_calendar.flat, expected)

    def test_overlap_intervals(self, dummy_calendar):
        dummy_calendar.add_intervals("precursor", "10d", gap="-5d")
        dummy_calendar = dummy_calendar.map_years(2021, 2021)
        expected = np.array(
            [
                interval("2021-12-16", "2021-12-26", closed="left"),
                interval("2021-12-21", "2021-12-31", closed="left"),
                interval("2021-12-31", "2022-01-20", closed="left"),
            ]
        )
        assert np.array_equal(dummy_calendar.flat, expected)

    def test_map_to_data(self, dummy_calendar):
        # create dummy data for testing
        time_index = pd.date_range("20201110", "20211211", freq="10d")
        var = np.random.random(len(time_index))
        # generate input data
        test_data = pd.Series(var, index=time_index)
        # map year to data
        calendar = dummy_calendar.map_to_data(test_data)
        # expected intervals
        expected = np.array(
            [
                interval("2020-12-21", "2020-12-31", closed="left"),
                interval("2020-12-31", "2021-01-20", closed="left"),
            ]
        )
        assert np.array_equal(calendar.flat, expected)

    def test_non_day_interval_length(self):
        cal = Calendar(anchor="December")
        cal.add_intervals("target", "1M")
        cal.add_intervals("precursor", "10M")
        cal.map_years(2020, 2020)
        expected = np.array(
            [
                interval("2020-02-01", "2020-12-01", closed="left"),
                interval("2020-12-01", "2021-01-01", closed="left"),
            ]
        )
        assert np.array_equal(cal.flat, expected)

    @pytest.mark.parametrize(
        "allow_overlap, expected_anchors",
        ((True, [2022, 2021, 2020]), (False, [2022, 2020])),
    )
    def test_allow_overlap(self, allow_overlap, expected_anchors):
        cal = Calendar(anchor="12-31", allow_overlap=allow_overlap)
        cal.add_intervals("target", length="30d")
        cal.add_intervals("precursor", length="365d")
        cal.map_years(2020, 2022)
        assert np.array_equal(expected_anchors, cal.get_intervals().index.values)


class TestAnchorKwarg:
    correct_inputs = [  # format: (anchor, anchor_fmt, anchor_str)
        ("5-5", "%m-%d", "5-5"),
        ("02-5", "%m-%d", "02-5"),
        ("06-05", "%m-%d", "06-05"),
        ("5-05", "%m-%d", "5-05"),
        ("11-30", "%m-%d", "11-30"),
        ("12-31", "%m-%d", "12-31"),
        ("W01", "W%W-%w", "W01-1"),
        ("W9", "W%W-%w", "W9-1"),
        ("W01-4", "W%W-%w", "W01-4"),
        ("W9-1", "W%W-%w", "W9-1"),
        ("December", "%m", "12"),
        ("dec", "%m", "12"),
        ("jan", "%m", "1"),
        ("Jan", "%m", "1"),
    ]

    incorrect_inputs = (
        10,  # non-str inputs
        (1, 2),
        "0",  # month number less than 1
        "13",  # month number greater than 12
        "12-0",  # day number less than 1
        "12-32",  # day number greater than 31
        "31-12",  # incorrect month/day order
        "4-31",  # 31 April (nonexistant date)
        "2-29",  # 29 February (doesn't exist in every year)
        "W60",  # Week number greater than 52 (only some years have 53 weeks)
        "W53",
        "W01-0",  # Weekday smaller than 1 (Monday)
        "W01-8",  # Weekday greater than 7 (Sunday)
        "w12",  # Small letter `w`.
        "juli",  # Non-English month name
        "July 5",  # Month name + day
        "July-5",
        "jan 20",
        "jan-20",
    )

    @pytest.mark.parametrize("test_input, expected_fmt, expected_str", correct_inputs)
    def test_correct_anchor_input(self, test_input, expected_fmt, expected_str):
        cal = Calendar(anchor=test_input)
        assert cal._anchor_fmt == expected_fmt  # pylint: disable=protected-access
        assert cal._anchor == expected_str  # pylint: disable=protected-access

    @pytest.mark.parametrize("test_input", incorrect_inputs)
    def test_incorrect_anchor_input(self, test_input):
        with pytest.raises(ValueError):
            _ = Calendar(anchor=test_input)


class TestMap:
    """Test map to year(s)/data methods"""

    def test_map_years(self):
        cal = daily_calendar(anchor="12-31", length="180d")
        cal.map_years(2020, 2021)
        expected = np.array(
            [
                [
                    interval("2021-07-04", "2021-12-31"),
                    interval("2021-12-31", "2022-06-29"),
                ],
                [
                    interval("2020-07-04", "2020-12-31"),
                    interval("2020-12-31", "2021-06-29"),  # notice the leap day
                ],
            ]
        )
        assert np.array_equal(cal.get_intervals(), expected)

    def test_map_years_single(self):
        cal = daily_calendar(anchor="12-31", length="180d")
        cal.map_years(2020, 2020)
        expected = np.array(
            [
                [
                    interval("2020-07-04", "2020-12-31"),
                    interval("2020-12-31", "2021-06-29"),
                ]
            ]
        )
        assert np.array_equal(cal.get_intervals(), expected)

    def test_map_to_data_edge_case_last_year(self):
        # test the edge value when the input could not cover the anchor date
        cal = daily_calendar(anchor="10-15", length="180d")
        # single year covered
        time_index = pd.date_range("20191020", "20211001", freq="60d")
        test_data = np.random.random(len(time_index))
        timeseries = pd.Series(test_data, index=time_index)
        cal.map_to_data(timeseries)
        expected = np.array(
            [
                [
                    interval("2020-04-18", "2020-10-15"),
                    interval("2020-10-15", "2021-04-13"),
                ]
            ]
        )
        assert np.array_equal(cal.get_intervals(), expected)

    def test_map_to_data_single_year_coverage(self):
        # test the single year coverage
        cal = daily_calendar(anchor="6-30", length="180d")
        # multiple years covered
        time_index = pd.date_range("20210101", "20211231", freq="7d")
        test_data = np.random.random(len(time_index))
        timeseries = pd.Series(test_data, index=time_index)
        cal.map_to_data(timeseries)

        expected = np.array(
            [
                [
                    interval("2021-01-01", "2021-06-30"),
                    interval("2021-06-30", "2021-12-27"),
                ]
            ]
        )

        assert np.array_equal(cal.get_intervals(), expected)

    def test_map_to_data_edge_case_first_year(self):
        # test the edge value when the input covers the anchor date
        cal = daily_calendar(anchor="10-15", length="180d")
        # multiple years covered
        time_index = pd.date_range("20191010", "20211225", freq="60d")
        test_data = np.random.random(len(time_index))
        timeseries = pd.Series(test_data, index=time_index)
        cal.map_to_data(timeseries)

        expected = np.array(
            [
                [
                    interval("2020-04-18", "2020-10-15"),
                    interval("2020-10-15", "2021-04-13"),
                ],
                [
                    interval("2019-04-18", "2019-10-15"),
                    interval("2019-10-15", "2020-04-12"),  # notice the leap day
                ],
            ]
        )

        assert np.array_equal(cal.get_intervals(), expected)

    def test_map_to_data_input_time_backward(self):
        # test when the input data has reverse order time index
        cal = daily_calendar(anchor="10-15", length="180d")
        time_index = pd.date_range("20201010", "20211225", freq="60d")
        test_data = np.random.random(len(time_index))
        timeseries = pd.Series(test_data, index=time_index[::-1])
        cal.map_to_data(timeseries)

        expected = np.array(
            [
                [
                    interval("2020-04-18", "2020-10-15"),
                    interval("2020-10-15", "2021-04-13"),
                ]
            ]
        )

        assert np.array_equal(cal.get_intervals(), expected)

    def test_map_to_data_xarray_input(self):
        # test when the input data has reverse order time index
        cal = daily_calendar(anchor="10-15", length="180d")
        time_index = pd.date_range("20201010", "20211225", freq="60d")
        test_data = np.random.random(len(time_index))
        dataarray = xr.DataArray(data=test_data, coords={"time": time_index})
        cal.map_to_data(dataarray)

        expected = np.array(
            [
                interval("2020-04-18", "2020-10-15"),
                interval("2020-10-15", "2021-04-13"),
            ]
        )

        assert np.all(cal.get_intervals() == expected)

    def test_missing_time_dim(self):
        cal = daily_calendar(anchor="10-15", length="180d")
        time_index = pd.date_range("20191020", "20211001", freq="60d")
        test_data = np.random.random(len(time_index))
        dataframe = pd.DataFrame(test_data, index=time_index)
        dataset = dataframe.to_xarray()
        with pytest.raises(ValueError):
            cal.map_to_data(dataset)

    def test_non_time_dim(self):
        cal = daily_calendar(anchor="10-15", length="180d")
        time_index = pd.date_range("20191020", "20211001", freq="60d")
        test_data = np.random.random(len(time_index))
        dataframe = pd.DataFrame(test_data, index=time_index)
        dataset = dataframe.to_xarray().rename({"index": "time"})
        dataset["time"] = np.arange(dataset["time"].size)
        with pytest.raises(ValueError):
            cal.map_to_data(dataset)

    # Note: add more test cases for different number of target periods!
    max_lag_edge_cases = [(73, [2019], 74), (72, [2019, 2018], 73)]
    # Test the edge cases of max_lag; where the max_lag just fits in exactly 365 days,
    # and where the max_lag just causes the calendar to skip a year

    @pytest.mark.parametrize("max_lag,expected_index,expected_size", max_lag_edge_cases)
    def test_max_lag_skip_years(self, max_lag, expected_index, expected_size):
        calendar = daily_calendar(anchor="12-31", length="5d", n_precursors=max_lag)
        calendar = calendar.map_years(2018, 2019)

        np.testing.assert_array_equal(
            calendar.get_intervals().index.values, expected_index
        )
        assert calendar.get_intervals().iloc[0].size == expected_size



"""s2spy train/test splitting methods.

Wrapper around sklearn splitters for working with (multiple) xarray dataarrays.
"""
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union
import numpy as np
import xarray as xr
from sklearn.model_selection._split import BaseCrossValidator
from sklearn.model_selection._split import BaseShuffleSplit


# Mypy type aliases
XType = Union[xr.DataArray, List[xr.DataArray]]
CVtype = Union[BaseCrossValidator, BaseShuffleSplit]

# For output types, variables are split in 2
XOnly = Tuple[XType, XType]
XAndY = Tuple[XType, XType, xr.DataArray, xr.DataArray]
XMaybeY = Iterable[Union[XOnly, XAndY]]


class CoordinateMismatchError(Exception):
    """Custom exception for unmatching coordinates."""


def _all_equal(arrays):
    """Return true if all arrays are equal."""
    try:
        arrays = iter(arrays)
        first = next(arrays)
        return all(np.array_equal(first, rest) for rest in arrays)
    except StopIteration:
        return True


class TrainTestSplit:
    """Split (multiple) xr.DataArrays across a given dimension."""

    def __init__(self, splitter: Type[CVtype]) -> None:
        """Split (multiple) xr.DataArrays across a given dimension.

        Calling `split()` on this object returns an iterator that allows passing in
        multiple input arrays at once. They need to have matching coordinates along
        the given dimension.

        For an overview of the sklearn Splitter Classes see:
        https://scikit-learn.org/stable/modules/classes.html#module-sklearn.model_selection

        Args:
            splitter (SplitterClass): Initialized splitter class, much have a
                `fit(X)` method which splits up `X` into multiple folds of
                train/test data.
        """
        self.splitter = splitter

    def split(
        self,
        *x_args: xr.DataArray,
        y: Optional[xr.DataArray] = None,
        dim: str = "anchor_year",
    ) -> XMaybeY:
        """Iterate over splits.

        Args:
            x_args: one or multiple xr.DataArray's that share the same
                coordinate along the given dimension
            y: (optional) xr.DataArray that shares the same coordinate along the
                given dimension
            dim: name of the dimension along which to split the data.

        Returns:
            Iterator over the splits
        """
        # Check that all inputs share the same dim coordinate
        coords = []
        x: xr.DataArray  # Initialize x to set scope outside loop
        for x in x_args:
            try:
                coords.append(x[dim])
            except KeyError as err:
                raise CoordinateMismatchError(
                    f"Not all input data arrays have the {dim} dimension."
                ) from err

        if not _all_equal(coords):
            raise CoordinateMismatchError(
                f"Input arrays are not equal along {dim} dimension."
            )

        if y is not None and not np.array_equal(y[dim], x[dim]):
            raise CoordinateMismatchError(
                f"Input arrays are not equal along {dim} dimension."
            )

        if x[dim].size <= 1:
            raise ValueError(
                f"Invalid input: need at least 2 values along dimension {dim}"
            )

        # Now we know that all inputs are equal.
        for train_indices, test_indices in self.splitter.split(x[dim]):
            if len(x_args) == 1:
                x_train: XType = x.isel({dim: train_indices})
                x_test: XType = x.isel({dim: test_indices})
            else:
                x_train = [da.isel({dim: train_indices}) for da in x_args]
                x_test = [da.isel({dim: test_indices}) for da in x_args]

            if y is None:
                yield x_train, x_test
            else:
                y_train = y.isel({dim: train_indices})
                y_test = y.isel({dim: test_indices})
                yield x_train, x_test, y_train, y_test



"""Tests for lilio's resample module.
"""
import tempfile
from pathlib import Path
import numpy as np
import pandas as pd
import pytest
import xarray as xr
from lilio import Calendar
from lilio import daily_calendar
from lilio import monthly_calendar
from lilio import resample
from lilio.resampling import VALID_METHODS


class TestResample:
    """Test resampling, general tests for how=mean."""

    # Define all required inputs as fixtures:
    @pytest.fixture
    def dummy_calendar(self):
        return daily_calendar(anchor="10-15", length="180d")

    @pytest.fixture(params=[1, 2])
    def dummy_calendar_targets(self, request):
        return daily_calendar(anchor="5-10", length="120d", n_targets=request.param)

    @pytest.fixture(params=["20151020", "20191015"])
    def dummy_series(self, request):
        time_index = pd.date_range(request.param, "20211001", freq="60d")
        test_data = np.random.random(len(time_index))
        expected = np.array([test_data[4:7].mean(), test_data[7:10].mean()])
        series = pd.Series(test_data, index=time_index, name="data1")
        return series, expected

    @pytest.fixture
    def dummy_dataframe(self, dummy_series):
        series, expected = dummy_series
        return pd.DataFrame(series), expected

    @pytest.fixture
    def dummy_dataarray(self, dummy_series):
        series, expected = dummy_series
        dataarray = series.to_xarray()
        dataarray = dataarray.rename({"index": "time"})
        return dataarray, expected

    @pytest.fixture
    def dummy_dataset(self, dummy_dataframe):
        dataframe, expected = dummy_dataframe
        dataset = dataframe.to_xarray().rename({"index": "time"})
        return dataset, expected

    @pytest.fixture
    def dummy_multidimensional(self):
        np.random.seed(0)
        time_index = pd.date_range("20171020", "20211001", freq="15d")
        return xr.Dataset(
            data_vars={
                "temp": (["x", "y", "time"], np.random.randn(2, 2, len(time_index))),
                "prec": (["x", "y", "time"], np.random.rand(2, 2, len(time_index))),
            },
            coords={
                "lon": (["x", "y"], [[-99.83, -99.32], [-99.79, -99.23]]),
                "lat": (["x", "y"], [[42.25, 42.21], [42.63, 42.59]]),
                "time": time_index,
            },
        )

    # Tests start here:
    def test_non_mapped_calendar(self, dummy_calendar):
        with pytest.raises(ValueError):
            resample(dummy_calendar, None)  # type: ignore

    def test_nontime_index(self, dummy_calendar, dummy_series):
        series, _ = dummy_series
        cal = dummy_calendar.map_to_data(series)
        series = series.reset_index()
        with pytest.raises(ValueError):
            resample(cal, series)

    def test_series(self, dummy_calendar, dummy_series):
        series, expected = dummy_series
        cal = dummy_calendar.map_to_data(series)
        resampled_data = resample(cal, series)
        np.testing.assert_allclose(resampled_data["data1"].iloc[:2], expected)

    def test_unnamed_series(self, dummy_calendar, dummy_series):
        series, expected = dummy_series
        series.name = None
        cal = dummy_calendar.map_to_data(series)
        resampled_data = resample(cal, series)
        np.testing.assert_allclose(resampled_data["data"].iloc[:2], expected)

    def test_dataframe(self, dummy_calendar, dummy_dataframe):
        dataframe, expected = dummy_dataframe
        cal = dummy_calendar.map_to_data(dataframe)
        resampled_data = resample(cal, dataframe)
        np.testing.assert_allclose(resampled_data["data1"].iloc[:2], expected)

    def test_dataarray(self, dummy_calendar, dummy_dataarray):
        dataarray, expected = dummy_dataarray
        cal = dummy_calendar.map_to_data(dataarray)
        resampled_data = resample(cal, dataarray)
        testing_vals = resampled_data.isel(anchor_year=0)
        np.testing.assert_allclose(testing_vals, expected)

    def test_dataset(self, dummy_calendar, dummy_dataset):
        dataset, expected = dummy_dataset
        cal = dummy_calendar.map_to_data(dataset)
        resampled_data = resample(cal, dataset)
        testing_vals = resampled_data["data1"].isel(anchor_year=0)
        np.testing.assert_allclose(testing_vals, expected)

    def test_multidim_dataset(self, dummy_calendar, dummy_multidimensional):
        cal = dummy_calendar.map_to_data(dummy_multidimensional)
        resampled_data = resample(cal, dummy_multidimensional)
        assert np.all([dim in resampled_data.dims for dim in ["x", "y"]])
        assert np.all([var in resampled_data.variables for var in ["temp", "prec"]])

    def test_target_period_dataframe(self, dummy_calendar_targets, dummy_dataframe):
        df, _ = dummy_dataframe
        calendar = dummy_calendar_targets.map_to_data(df)
        resampled_data = resample(calendar, df)
        expected = np.zeros(resampled_data.index.size, dtype=bool)
        for i in range(calendar.n_targets):
            expected[i::3] = True
        np.testing.assert_array_equal(
            resampled_data["is_target"].values, expected[::-1]
        )

    def test_target_period_dataset(self, dummy_calendar_targets, dummy_dataset):
        ds, _ = dummy_dataset
        calendar = dummy_calendar_targets.map_to_data(ds)
        resampled_data = resample(calendar, ds)
        expected = np.zeros(3, dtype=bool)
        expected[: dummy_calendar_targets.n_targets] = True
        np.testing.assert_array_equal(
            resampled_data["is_target"].values, expected[::-1]
        )

    def test_allow_overlap_dataframe(self):
        calendar = daily_calendar(
            anchor="10-15", length="100d", n_precursors=5, allow_overlap=True
        )

        time_index = pd.date_range("20151101", "20211101", freq="50d")
        test_data = np.random.random(len(time_index))
        series = pd.Series(test_data, index=time_index)
        calendar.map_to_data(series)
        intervals = calendar.get_intervals()
        # 4 anchor years are expected if overlap is allowed
        assert len(intervals.index) == 4

    def test_1day_freq_dataframe(self):
        # Will test the regular expression match and pre-pending of '1' in the
        # check_input_frequency utility function
        calendar = daily_calendar(anchor="10-15", length="1d")
        time_index = pd.date_range("20191101", "20211101", freq="1d")
        test_data = np.random.random(len(time_index))
        series = pd.Series(test_data, index=time_index, name="data1")
        calendar.map_to_data(series)
        calendar.get_intervals()

    def test_to_netcdf(self, dummy_calendar, dummy_dataset):
        # Test to ensure that xarray data resampled using the calendar can be written
        # to a netCDF file.
        dataset, _ = dummy_dataset
        cal = dummy_calendar.map_to_data(dataset)
        resampled_data = resample(cal, dataset)
        with tempfile.TemporaryDirectory() as tmpdirname:
            path = Path(tmpdirname) / "test.nc"
            resampled_data.to_netcdf(path)

    def test_overlapping(self):
        # Test to ensure overlapping intervals are accepted and correctly resampled
        time_index = pd.date_range("20191001", "20200101", freq="30d")
        test_data = np.random.random(len(time_index))
        series = pd.Series(test_data, index=time_index, name="data1")

        calendar = Calendar(anchor="10-05")
        calendar.add_intervals("target", "60d")
        calendar.add_intervals("precursor", "60d")
        calendar.add_intervals("precursor", "60d", gap="-60d")

        calendar.map_to_data(series)
        resampled_data = resample(calendar, series)

        expected = np.array(
            [
                np.mean(series.values[-5:-3]),
                np.mean(series.values[-5:-3]),
                np.mean(series.values[-3:-1]),
            ]
        )

        np.testing.assert_array_equal(resampled_data["data1"].values[-3:], expected)

    # Test data for missing intervals, too low frequency.
    def test_missing_intervals_dataframe(self, dummy_calendar, dummy_dataframe):
        dataframe, _ = dummy_dataframe
        cal = dummy_calendar.map_years(2020, 2025)
        with pytest.warns(UserWarning):
            resample(cal, dataframe)

    def test_missing_intervals_dataset(self, dummy_calendar, dummy_dataset):
        dataset, _ = dummy_dataset
        cal = dummy_calendar.map_years(2020, 2025)
        with pytest.warns(UserWarning):
            resample(cal, dataset)


TOO_LOW_FREQ_ERR = r".*lower time resolution than the calendar.*"
TOO_LOW_FREQ_WARN = r".*input data frequency is very close to the Calendar's freq.*"


class TestResampleChecks:
    @pytest.fixture
    def dummy_dataframe(self):
        time_index = pd.date_range("20191015", "20211001", freq="2d")
        test_data = np.random.random(len(time_index))
        series = pd.Series(test_data, index=time_index, name="data1")
        return pd.DataFrame(series)

    @pytest.fixture
    def dummy_dataset(self, dummy_dataframe):
        return dummy_dataframe.to_xarray().rename({"index": "time"})

    def test_low_freq_warning_dataframe(self, dummy_dataframe):
        cal = daily_calendar(anchor="10-15", length="2d")
        cal = cal.map_to_data(dummy_dataframe)
        with pytest.warns(UserWarning, match=TOO_LOW_FREQ_WARN):
            resample(cal, dummy_dataframe)

    def test_too_low_freq_dataframe(self, dummy_dataframe):
        cal = daily_calendar(anchor="10-15", length="1d")
        cal = cal.map_to_data(dummy_dataframe)
        with pytest.raises(ValueError, match=TOO_LOW_FREQ_ERR):
            resample(cal, dummy_dataframe)

    def test_low_freq_warning_dataset(self, dummy_dataset):
        cal = daily_calendar(anchor="10-15", length="2d")
        cal = cal.map_to_data(dummy_dataset)
        with pytest.warns(UserWarning, match=TOO_LOW_FREQ_WARN):
            resample(cal, dummy_dataset)

    def test_too_low_freq_dataset(self, dummy_dataset):
        cal = daily_calendar(anchor="10-15", length="1d")
        cal = cal.map_to_data(dummy_dataset)
        with pytest.raises(ValueError, match=TOO_LOW_FREQ_ERR):
            resample(cal, dummy_dataset)

    def test_low_freq_month_fmt_dataframe(self):
        time_index = pd.date_range("20181001", "20211001", freq="20d")
        df = pd.DataFrame(
            data={
                "data1": np.random.random(len(time_index)),
            },
            index=time_index,
        )
        cal = monthly_calendar(anchor="10-15", length="1M")
        cal = cal.map_to_data(df)
        with pytest.warns(UserWarning, match=TOO_LOW_FREQ_WARN):
            resample(cal, df)

    def test_month_freq_data(self):
        time_index = pd.date_range("20181001", "20211001", freq="2M")
        test_data = pd.DataFrame(
            data={
                "data1": np.random.random(len(time_index)),
            },
            index=time_index,
        )
        cal = monthly_calendar(anchor="Dec", length="1M")
        cal.map_to_data(test_data)
        with pytest.raises(ValueError, match=TOO_LOW_FREQ_ERR):
            resample(cal, test_data)

    def test_reserved_names_dataframe(self, dummy_dataframe):
        cal = daily_calendar(anchor="10-15", length="7d")
        cal.map_to_data(dummy_dataframe)
        with pytest.raises(ValueError, match=r".*reserved names..*"):
            resample(cal, dummy_dataframe.rename(columns={"data1": "anchor_year"}))


class TestResampleMethods:
    """Test alternative resampling methods.

    Note: outcomes are not tested. The outcome of np.mean is tested above in
    TestResample. If those tests pass fine with np.mean being used through
    argparse, then these should be correct as well."""

    @pytest.fixture
    def dummy_calendar(self):
        return daily_calendar(anchor="10-15", length="180d")

    @pytest.fixture
    def dummy_dataframe(self):
        time_index = pd.date_range("20151020", "20211001", freq="60d")
        test_data = np.random.random(len(time_index))
        expected = np.array([test_data[4:7].mean(), test_data[7:10].mean()])
        series = pd.Series(test_data, index=time_index, name="data1")
        return pd.DataFrame(series), expected

    @pytest.fixture
    def dummy_dataset(self, dummy_dataframe):
        dataframe, expected = dummy_dataframe
        dataset = dataframe.to_xarray().rename({"index": "time"})
        return dataset, expected

    @pytest.mark.parametrize("resampling_method", VALID_METHODS)
    def test_all_methods_dataframe(
        self, dummy_calendar, dummy_dataframe, resampling_method
    ):
        data, _ = dummy_dataframe
        cal = dummy_calendar.map_to_data(data)
        resample(cal, data, how=resampling_method)

    @pytest.mark.parametrize("resampling_method", VALID_METHODS)
    def test_all_methods_dataset(
        self, dummy_calendar, dummy_dataset, resampling_method
    ):
        data, _ = dummy_dataset
        cal = dummy_calendar.map_to_data(data)
        resample(cal, data, how=resampling_method)

    def test_func_input_dataframe(self, dummy_calendar, dummy_dataframe):
        data, _ = dummy_dataframe
        cal = dummy_calendar.map_to_data(data)
        resample(cal, data, how=np.mean)

    def test_func_input_dataset(self, dummy_calendar, dummy_dataset):
        data, _ = dummy_dataset
        cal = dummy_calendar.map_to_data(data)
        resample(cal, data, how=np.mean)



"""Shorthands for calendars, to make generating commonly used calendars a one-liner."""
import re
import pandas as pd
from .calendar import Calendar


def daily_calendar(
    anchor: str,
    length: str = "1d",
    n_targets: int = 1,
    n_precursors: int = 0,
    allow_overlap: bool = False,
) -> Calendar:
    """Instantiate a basic daily calendar with minimal configuration.

    Set up a quick calendar revolving around intervals with day-based lengths.
    The intervals will extend back in time with as many intervals as fit within the
    cycle time of one year.

    Args:
        anchor: String in the form "12-31" for December 31st. The first target interval
            will contain the anchor, while the precursor intervals are built back in
            time starting at this date.
        length: The length of every target and precursor period.
        n_targets: integer specifying the number of target intervals in a period.
        n_precursors: Sets the maximum number of precursors of the Calendar. If
            `0`, the amount will be determined by how many fit in each anchor year.
            If a value is provided, the intervals can either only cover part
            of the year, or extend over multiple years. In case of a large max_lag
            number where the intervals extend over multiple years, anchor years will
            be skipped to avoid overlapping intervals. To allow overlapping
            intervals, use the `allow_overlap` kwarg.
        allow_overlap: Allows intervals to overlap between anchor years, if the
            max_lag is set to a high enough number that intervals extend over
            multiple years. `False` by default, to avoid train/test information
            leakage.

    Returns:
        An instantiated Calendar built according to the input kwarg specifications

    Example:
        Instantiate a calendar counting towards Christmas in 3-days steps.

        >>> import lilio
        >>> calendar = lilio.daily_calendar(anchor='12-25', length="3d", n_precursors=3)
        >>> calendar  # doctest: +NORMALIZE_WHITESPACE
        Calendar(
            anchor='12-25',
            allow_overlap=False,
            mapping=None,
            intervals=[
                Interval(role='target', length='3d', gap='0d'),
                Interval(role='precursor', length='3d', gap='0d'),
                Interval(role='precursor', length='3d', gap='0d'),
                Interval(role='precursor', length='3d', gap='0d')
            ]
        )

    """
    if not re.fullmatch(r"\d*d", length):
        raise ValueError("Please input a frequency in the form of '2d'")
    periods_per_year = pd.Timedelta("365days") / pd.to_timedelta(length)
    n_intervals = (
        (n_precursors + n_targets) if n_precursors > 0 else int(periods_per_year)
    )
    n_precursors = n_intervals - n_targets

    cal = Calendar(anchor=anchor, allow_overlap=allow_overlap)

    cal.add_intervals(role="target", length=length, n=n_targets)
    if n_precursors > 0:
        cal.add_intervals(role="precursor", length=length, n=n_precursors)

    return cal


def weekly_calendar(
    anchor: str,
    length: str = "1W",
    n_targets: int = 1,
    n_precursors: int = 0,
    allow_overlap: bool = False,
) -> Calendar:
    """Instantiate a basic monthly calendar with minimal configuration.

    Set up a quick calendar revolving around intervals with week-based lengths.
    The precursor intervals will extend back in time with as many intervals as fit
    within the cycle time of one year (i.e. 52 - n_targets).

    Note that the difference between this calendar and the daily_calendar revolves
    around the use of calendar weeks (Monday - Sunday), instead of 7-day periods.

    Args:
        anchor: Str in the form of "40W", denoting the week number. The first target
            interval will contain the anchor, while the precursor intervals are built
            back in time starting from this week.
        length: The length of every precursor and target interval, e.g. '2W'.
        n_targets: integer specifying the number of target intervals in a period.
        n_precursors: Sets the maximum number of precursors of the Calendar. If
            `0`, the amount will be determined by how many fit in each anchor year.
            If a value is provided, the intervals can either only cover part
            of the year, or extend over multiple years. In case of a large max_lag
            number where the intervals extend over multiple years, anchor years will
            be skipped to avoid overlapping intervals. To allow overlapping
            intervals, use the `allow_overlap` kwarg.
        allow_overlap: Allows intervals to overlap between anchor years, if the
            max_lag is set to a high enough number that intervals extend over
            multiple years. `False` by default, to avoid train/test information
            leakage.

    Returns:
        An instantiated Calendar built according to the input kwarg specifications

    Example:
        Instantiate a calendar counting down the quarters (3 month periods) from
        december.

        >>> import lilio
        >>> calendar = lilio.weekly_calendar(anchor="W40", length="1W", n_precursors=2)
        >>> calendar  # doctest: +NORMALIZE_WHITESPACE
        Calendar(
            anchor='W40-1',
            allow_overlap=False,
            mapping=None,
            intervals=[
                Interval(role='target', length='1W', gap='0d'),
                Interval(role='precursor', length='1W', gap='0d'),
                Interval(role='precursor', length='1W', gap='0d')
            ]
        )

    """
    if not re.fullmatch(r"\d*W", length):
        raise ValueError("Please input a frequency in the form of '4W'")
    periods_per_year = pd.Timedelta("365days") / pd.to_timedelta(length)
    n_intervals = (
        (n_precursors + n_targets) if n_precursors > 0 else int(periods_per_year)
    )
    n_precursors = n_intervals - n_targets

    cal = Calendar(anchor=anchor, allow_overlap=allow_overlap)

    cal.add_intervals(role="target", length=length, n=n_targets)
    if n_precursors > 0:
        cal.add_intervals(role="precursor", length=length, n=n_precursors)

    return cal


def monthly_calendar(
    anchor: str,
    length: str = "1M",
    n_targets: int = 1,
    n_precursors: int = 0,
    allow_overlap: bool = False,
) -> Calendar:
    """Instantiate a basic monthly calendar with minimal configuration.

    Set up a quick calendar revolving around intervals with month-based lengths.
    The intervals will extend back in time with as many intervals as fit within the
    cycle time of one year.

    Args:
        anchor: Str in the form 'January' or 'Jan'. he first target interval
            will contain the anchor, while the precursor intervals are built back in
            time starting at this Month.
        length: The length of every target and precursor period, in the form '1M',
            '2M', etc.
        n_targets: integer specifying the number of target intervals in a period.
        n_precursors: Sets the maximum number of precursors of the Calendar. If
            `0`, the amount will be determined by how many fit in each anchor year.
            If a value is provided, the intervals can either only cover part
            of the year, or extend over multiple years. In case of a large max_lag
            number where the intervals extend over multiple years, anchor years will
            be skipped to avoid overlapping intervals. To allow overlapping
            intervals, use the `allow_overlap` kwarg.
        allow_overlap: Allows intervals to overlap between anchor years, if the
            max_lag is set to a high enough number that intervals extend over
            multiple years. `False` by default, to avoid train/test information
            leakage.

    Returns:
        An instantiated Calendar built according to the input kwarg specifications

    Example:
        Instantiate a calendar counting down the quarters (3 month periods) from
        december.

        >>> import lilio
        >>> calendar = lilio.monthly_calendar(anchor='Dec', length="3M")
        >>> calendar  # doctest: +NORMALIZE_WHITESPACE
        Calendar(
            anchor='12',
            allow_overlap=False,
            mapping=None,
            intervals=[
                Interval(role='target', length='3M', gap='0d'),
                Interval(role='precursor', length='3M', gap='0d'),
                Interval(role='precursor', length='3M', gap='0d'),
                Interval(role='precursor', length='3M', gap='0d')
            ]
        )

    """
    if not re.fullmatch(r"\d*M", length):
        raise ValueError("Please input a frequency in the form of '2M'")
    periods_per_year = 12 / int(length.replace("M", ""))
    n_intervals = (
        (n_precursors + n_targets) if n_precursors > 0 else int(periods_per_year)
    )
    n_precursors = n_intervals - n_targets

    cal = Calendar(anchor=anchor, allow_overlap=allow_overlap)

    cal.add_intervals(role="target", length=length, n=n_targets)
    if n_precursors > 0:
        cal.add_intervals(role="precursor", length=length, n=n_precursors)

    return cal
